from typing import List
from abc import ABCMeta, abstractmethod

from dailygoals_services.model import SaleChallenge


class DailyGoalsRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_sales_challenge(self, limit):
        # type: (int) -> List[SaleChallenge]

        raise NotImplementedError

    @abstractmethod
    def insert_sales_challenge(self, challenge):
        # type: (SaleChallenge) -> int

        raise NotImplementedError

    @abstractmethod
    def delete_sales_challenge(self, sale_challenge_id):
        # type: (int) -> int

        raise NotImplementedError

    @abstractmethod
    def update_sales_challenge(self, challenge):
        # type: (SaleChallenge) -> int

        raise NotImplementedError



def migrate(conn, tables):
    updateTablesPriority = ["ProductionOrders"]

    for table in updateTablesPriority:
        if len(tables[table]["fields_Common"]) == 0:
            continue
        strCommonFields = ", ".join(tables[table]["fields_Common"])
        stmt = "INSERT OR REPLACE INTO %s (%s) SELECT %s FROM old.%s;" % (table, strCommonFields, strCommonFields, table)
        conn.query(stmt)



from typing import Optional


class TablePicture(object):
    def __init__(
        self,
        store_id,           # type: str
        table_info=None,    # type: Optional[dict]
    ):
        self.store_code = store_id
        self.table_info = table_info



import json

from messagebus import AckMessage

from ._CommonProcessor import _CommonProcessor


class GetElectronicTypesProcessor(_CommonProcessor):
    def parse_data(self, data, event_type):
        return data

    def call_business(self, _):
        return self.payment_router.get_electronic_types()

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(message, AckMessage(json.dumps(result).encode("utf-8")))



from typing import Optional


class UnlockVoucherRequest(object):
    def __init__(
        self,               # type: UnlockVoucherRequest
        benefit_id,         # type: str
        order_picture,      # type: str
        table_picture,      # type: Optional[str]
    ):
        # type: (...) -> None

        self.benefit_id = benefit_id
        self.order_picture = order_picture
        self.table_picture = table_picture



# iban.py - functions for handling Belgian IBANs
# coding: utf-8
#
# Copyright (C) 2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Belgian IBAN (International Bank Account Number).

The IBAN is used to identify bank accounts across national borders. The
Belgian IBAN is built up of the IBAN prefix (BE) and check digits, followed
by a 3 digit bank identifier, a 7 digit account number and 2 more check
digits.

* https://www.nbb.be/en/payment-systems/payment-standards/bank-identification-codes

>>> validate('BE32 123-4567890-02')
'BE32123456789002'
>>> validate('BE41091811735141')  # incorrect national check digits
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('BE83138811735115')  # unknown bank code
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('GR1601101050000010547023795')  # not a Belgian IBAN
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> to_bic('BE 48 3200 7018 4927')
'BBRUBEBB'
>>> to_bic('BE83138811735115') is None
True
"""

from stdnum import iban
from stdnum.exceptions import *


__all__ = ['compact', 'format', 'validate', 'is_valid']


compact = iban.compact
format = iban.format


def _calc_check_digits(number):
    """Calculate the check digits over the provided part of the number."""
    check = int(number) % 97
    return '%02d' % (check or 97)


def info(number):
    """Return a dictionary of data about the supplied number. This typically
    returns the name of the bank and a BIC if it is valid."""
    number = compact(number)
    from stdnum import numdb
    return numdb.get('be/banks').info(number[4:7])[0][1]


def to_bic(number):
    """Return the BIC for the bank that this number refers to."""
    bic = info(number).get('bic')
    if bic:
        return str(bic)


def validate(number):
    """Check if the number provided is a valid Belgian IBAN."""
    number = iban.validate(number, check_country=False)
    if not number.startswith('BE'):
        raise InvalidComponent()
    if number[-2:] != _calc_check_digits(number[4:-2]):
        raise InvalidChecksum()
    if not info(number):
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number provided is a valid Belgian IBAN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from application.interactor import MediaUpdaterInteractor
from application.manager._UpdaterManager import UpdaterManager
from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import NoUpdateFound
from application.repository.pos import (
    MediaUpdaterPOSRepository,
    BinaryUpdaterPOSRepository,
)
from application.util import LoggerUtil

logger = LoggerUtil.get_logger_name(__name__)


class MediaUpdaterManager(UpdaterManager):
    def __init__(
        self,
        configs,                # type: Configurations
        interactor,             # type: MediaUpdaterInteractor
        pos_repository,         # type: MediaUpdaterPOSRepository
        binaries_repository,    # type: BinaryUpdaterPOSRepository
    ):
        # type: (...) -> None # noqa

        name = UpdateType.media.name
        super(MediaUpdaterManager, self).__init__(name, configs, interactor, pos_repository, logger)
        self.interactor = interactor
        self.binaries_repository = binaries_repository

        self.logger = logger

    def run(self, validate_window=True):
        # type: (bool) -> None

        super(MediaUpdaterManager, self).run(False)

    def update(self):
        # type: () -> None

        self.interactor.update_media()

    def has_update_to_apply(self):
        # type: () -> None

        if not self.binaries_repository.has_already_applied_update():
            raise NoUpdateFound()

        if not self.pos_repository.has_pending_update_to_apply():
            raise NoUpdateFound()

        if not self.pos_repository.has_pending_update_to_apply(UpdateType.banners.value):
            raise NoUpdateFound()



import sys
from abc import ABCMeta
from logging import Logger

from production.box.statistic import StatisticBox
from production.manager import StatisticUpdater
from production.treebuilder import BoxConfiguration
from typing import Dict

from ._BoxBuilder import BoxBuilder


class StatisticBoxBuilder(BoxBuilder):
    __metaclass__ = ABCMeta

    def __init__(self, statistic_updater, loggers):
        # type: (StatisticUpdater, Dict[str, Logger]) -> None
        super(StatisticBoxBuilder, self).__init__(loggers)
        self.statistic_updater = statistic_updater

    def build(self, config):
        # type: (BoxConfiguration) -> StatisticBox

        def get_config(config_name):
            return self.get_extra_config(config, config_name)

        self.loggers['DEBUG'].info('creating statistic box...')
        statistic_class_name = self.get_extra_config(config, "StatisticClass")

        statistic_type = getattr(sys.modules["production.box.statistic"], statistic_class_name, None)
        if statistic_type is None:
            self.loggers['DEBUG'].info("StatisticClass ({}) does not exist".format(statistic_class_name))
            raise Exception("StatisticClass ({}) does not exist".format(statistic_class_name))

        self.loggers['DEBUG'].info('StatisticClass = {}'.format(statistic_class_name))
        self.loggers['DEBUG'].info('StatisticType = {}'.format(statistic_type))
        return statistic_type(config.name, config.sons, self.statistic_updater, self.get_logger(config), get_config)

    def get_extra_config(self, config, config_name):
        config_value = self.get_extra(config, config_name, None)
        if config_value is None:
            self.loggers['DEBUG'].info("({}) does not exist".format(config_name))
            raise Exception("({}) does not exist".format(config_name))

        return config_value



class LogisticPartnerNotFound(Exception):
    pass


class LogisticSearchError(Exception):
    pass


class LogisticCancelError(Exception):
    pass


class LogisticRequestError(Exception):
    pass


class SaveDeliveryOrderError(Exception):
    pass


class SendLogisticError(Exception):
    pass


class SendOrderToProductionError(Exception):
    pass


class DeliveryCouponPrintError(Exception):
    pass


class NotFoundUserError(Exception):
    pass


class GettingPosVersionException(Exception):
    pass


class BenefitAppliedException(Exception):
    pass


class RemoveBenefitException(Exception):
    pass


class ErrorGettingBenefits(Exception):
    pass


class InvalidTipError(Exception):

    class ErrorCode(object):
        CANNOT_USE_TIP_BY_AMOUNT_AND_PERCENTAGE = 1
        TIP_AMOUNT_MUST_BE_VALID = 2

    def __init__(
        self,       # type: InvalidTipError
        code,       # type: InvalidTipError.ErrorCode
    ):
        # type: (...) -> None
        super(InvalidTipError, self).__init__()

        self.code = code


class UpdateRuptureException(Exception):
    pass


class CleanRuptureException(Exception):
    pass


class ErrorGettingSalesChallenge(Exception):
    pass


class ErrorOnInsertSalesChallenge(Exception):
    pass


class ErrorOnUpdateSalesChallenge(Exception):
    pass


class ErrorOnDeleteSalesChallenge(Exception):
    pass


class DateFormatException(Exception):
    pass


class GettingUserInfoException(Exception):
    pass


class CanceledDialogException(Exception):
    pass


class RequiredDocumentException(Exception):
    pass


class ErrorGettingNFType(Exception):
    pass


class ErrorGettingCertificateExpirationDays(Exception):
    pass


class ErrorApplyingDiscount(Exception):
    def __init__(
        self,        # type: ErrorApplyingDiscount
        error_code,  # type: str
    ):
        # type: (...) -> None

        self.error_code = error_code


class DiscountMinLevelException(Exception):
    def __init__(
        self,
        min_level_required,     # type: int
    ):
        # type: (...) -> None

        self.min_level_required = min_level_required


class GetRuptureItemsException(Exception):
    pass


class MinQuantityException(Exception):
    pass


class MaxQuantityException(Exception):
    pass


class ElectronicPaymentException(Exception):
    pass


class ProductNotExistsException(Exception):
    pass


class MissingTenderParametersException(Exception):
    pass


class RemoteOrderStoreStatusError(Exception):
    pass


class ErrorGettingDevicesData(Exception):
    pass


class ErrorGettingProductData(Exception):
    pass


class ComponentNotFound(Exception):
    def __init__(
        self,
        component,  # type: str
    ):
        # type: (...) -> None

        content = "$COMPONENT_NOT_FOUND|{}".format(component)
        super(ComponentNotFound, self).__init__(content)


class PrinterError(Exception):
    def __init__(self, message):
        super(PrinterError, self).__init__(message)


class ErrorGettingDevice(Exception):
    def __init__(
        self,
        pos_id,             # type: str
        device_type,        # type: str
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.device_type = device_type


class CloseOrderByOperatorCloseException(Exception):
    def __init__(
        self,
        service_type,   # type: str
        pos_id,         # type: str
    ):
        # type: (...) -> None

        content = "Operation storing {} order cancelled by operator in POS: {}".format(service_type, pos_id)
        super(CloseOrderByOperatorCloseException, self).__init__(content)


class OrderNotFoundException(Exception):
    def __init__(
        self,           # type: OrderNotFoundException
        error_code,     # type: str
        message,        # type: str
    ):
        # type: (...) -> None

        content = "$ERROR_CODE_INFO|{}|{}".format(
            error_code,
            message,
        )
        super(OrderNotFoundException, self).__init__(content)


class RestartComponentException(Exception):
    def __init__(
        self,
        component,  # type: str
        token,      # type: str
        data,       # type: str
    ):
        # type: (...) -> None

        content = "Error stopping component {0}. {1}: {2}".format(
            component,
            token,
            data,
        )
        super(ComponentNotFound, self).__init__(content)


class SystemUserAccessException(Exception):
    pass



# pan.py - functions for handling Indian income tax numbers
#
# Copyright (C) 2017 Srikanth Lakshmanan
# Copyright (C) 2021 Gaurav Chauhan
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""PAN (Permanent Account Number, Indian income tax identifier).

The Permanent Account Number (PAN) is a 10 digit alphanumeric identifier for
Indian individuals, families and corporates for income tax purposes. It is
also issued to foreign nationals subject to a valid visa.

PAN is made up of 5 letters, 4 digits and 1 alphabetic check digit. The 4th
character indicates the type of holder, the 5th character is either 1st
letter of the holder's name or holder's surname in case of 'Individual' PAN,
next 4 digits are serial numbers running from 0001 to 9999 and the last
character is a check digit computed by an undocumented checksum algorithm.

More information:

* https://en.wikipedia.org/wiki/Permanent_account_number
* https://incometaxindia.gov.in/tutorials/1.permanent%20account%20number%20(pan).pdf

>>> validate('ACUPA7085R')
'ACUPA7085R'
>>> validate('ACUPA7085RR')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('ABMPA32111')  # check digit should be a letter
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('ABMXA3211G')  # invalid type of holder
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('ACUPA0000R')  # serial number should not be '0000'
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> mask('AAPPV8261K')
'AAPPVXXXXK'
>>> info('AAPPV8261K')['holder_type']
'Individual'
"""

import re

from stdnum.exceptions import *
from stdnum.util import clean


_pan_re = re.compile(r'^[A-Z]{5}[0-9]{4}[A-Z]$')

_pan_holder_types = {
    'A': 'Association of Persons (AOP)',
    'B': 'Body of Individuals (BOI)',
    'C': 'Company',
    'F': 'Firm/Limited Liability Partnership',
    'G': 'Government Agency',
    'H': 'Hindu Undivided Family (HUF)',
    'L': 'Local Authority',
    'J': 'Artificial Juridical Person',
    'P': 'Individual',
    'T': 'Trust',
    'K': 'Krish (Trust Krish)',
}
# Type 'K' may have been discontinued, not listed on Income Text Dept website.


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').upper().strip()


def validate(number):
    """Check if the number provided is a valid PAN. This checks the length
    and formatting."""
    number = compact(number)
    if len(number) != 10:
        raise InvalidLength()
    if not _pan_re.match(number):
        raise InvalidFormat()
    info(number)  # used to check 4th digit
    if number[5:9] == '0000':
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number provided is a valid PAN. This checks the length
    and formatting."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def info(number):
    """Provide information that can be decoded from the PAN."""
    number = compact(number)
    holder_type = _pan_holder_types.get(number[3])
    if not holder_type:
        raise InvalidComponent()
    return {
        'holder_type': holder_type,
        'card_holder_type': holder_type,  # for backwards compatibility
        'initial': number[4],
    }


def mask(number):
    """Mask the PAN as per Central Board of Direct Taxes (CBDT) masking
    standard."""
    number = compact(number)
    return number[:5] + 'XXXX' + number[-1:]



from helper import ExtendedEnum


class GetAndLockCouponErrorCode(ExtendedEnum):
    COUPON_NOT_FOUND = "The voucher id was not found on the partner"
    PARTNER_CUSTOM_ERROR = "The voucher is not available based on custom partner rules"
    COUPON_LOCKED = "The voucher is being used in another pos"
    COUPON_ALREADY_USED = "The voucher was already used"



# rut.py - functions for handling Uruguay RUT numbers
# coding: utf-8
#
# Copyright (C) 2019 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""RUT (Registro Único Tributario, Uruguay tax number).

The Registro Único Tributario (RUT) is an identifier of legal entities for
tax purposes.

This number consists of 12 digits, the first two indicate the registration
number, followed by a 6 digit sequence number, followed by 001 and a check
digit.

More information:

* https://www.agesic.gub.uy/innovaportal/file/1634/1/modelo_de_datos.pdf (page 71)
* https://servicios.dgi.gub.uy/ServiciosEnLinea/dgi--servicios-en-linea--consulta-de-certifcado-unico

>>> validate('21-100342-001-7')
'211003420017'
>>> validate('UY 21 140634 001 1')
'211406340011'
>>> validate('210303670014')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('12345678')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('211003420017')
'21-100342-001-7'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


# There are various online validation services available but they require
# registration and WS-Securety signatures.
# https://www.agesic.gub.uy/innovaportal/v/1600/9/agesic/consulta-de-entidad-por-rut.html
# https://servicios.dgi.gub.uy/ServiciosEnLinea/ampliar/servicios-automatizados


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    number = clean(number, ' -').upper().strip()
    if number.startswith('UY'):
        return number[2:]
    return number


def calc_check_digit(number):
    """Calculate the check digit."""
    weights = (4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2)
    total = sum(int(n) * w for w, n in zip(weights, number))
    return str(-total % 11)


def validate(number):
    """Check if the number is a valid Uruguay RUT number.

    This checks the length, formatting and check digit.
    """
    number = compact(number)
    if len(number) != 12:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[:2] < '01' or number[:2] > '21':
        raise InvalidComponent()
    if number[2:8] == '000000':
        raise InvalidComponent()
    if number[8:11] != '001':
        raise InvalidComponent()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid Uruguay RUT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join([number[:2], number[2:-4], number[-4:-1], number[-1]])



# ytunnus.py - functions for handling Finnish business identifiers (y-tunnus)
# coding: utf-8
#
# Copyright (C) 2015 Holvi Payment Services Oy
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Y-tunnus (Finnish business identifier).

The number is an 8-digit code with a weighted checksum.

>>> validate('2077474-0')
'20774740'
>>> validate('2077474-1')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('2077474-0')
'2077474-0'
"""

from stdnum.exceptions import *
from stdnum.fi import alv


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return alv.compact(number)


def validate(number):
    """Check if the number is a valid business identifier. This checks the
    length, formatting and check digit."""
    return alv.validate(number)


def is_valid(number):
    """Check if the number is a valid business identifier."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return number[:7] + '-' + number[7:]



import logging

from typing import (
    Any,
    List,
    Dict,
    Optional,
    Tuple,
    Union
)
from xml.etree import cElementTree as eTree

from application.domain import TableService
from pos_model import TableStatus
from sysactions import (
    get_model,
    show_listbox,
    StopAction,
    translate_message,
)


class TableInteractor(object):
    def __init__(
            self,
            pos_id,         # type: str
            table_id,       # type: str
            table_service   # type: TableService
    ):
        # type: (...) -> None

        self.logger = logging.getLogger("PosActions")
        self.table_service = table_service
        self.pos_id = pos_id
        self.table_id = table_id

    def get_related_services(
        self,
        filter_totaled=False    # type: Optional[bool]
    ):
        # type: (...) -> Tuple[List[Dict], List[str]]

        table_picture = eTree.XML(
                self.table_service.get_table_picture(
                        self.pos_id,
                        self.table_id
                )
        )
        model = get_model(self.pos_id)
        related_objects = []
        services_totaled = []
        service_children = table_picture.findall(".//ServiceChildren/ServiceChild")
        for child in service_children:
            child_is_totaled = int(child.get("status") or 0) == TableStatus.TOTALED
            if child_is_totaled:
                services_totaled.append(child.get("tableId"))

            child_is_closed = int(child.get("status") or 0) in [TableStatus.CLOSED, TableStatus.ABANDONED]
            if child_is_closed or (filter_totaled and child_is_totaled):
                continue
            tab_label = ' / '.join([
                data
                for data in [
                    child.get("tabId"),
                    child.get("customerName"),
                    child.get("customerPhone")
                ]
                if data
            ])
            related_objects.append({
                "caption": translate_message(model, "TAB_INFORMATION", tab_label),
                "table_id": child.get("tableId"),
            })

        return related_objects, services_totaled

    def get_selected_related_services(
        self,
        related_objects,        # type: List[Dict]
        show_all_services=False    # type: bool
    ):
        # type: (...) -> str

        if len(related_objects) > 0:
            listbox_options = [e["caption"] for e in related_objects]
            if show_all_services:
                model = get_model(self.pos_id)
                listbox_options.append(translate_message(model, "CLOSE_ALL_TABLES_OR_TABS"))
                related_objects.append({
                    "table_id": "all_services",
                })
            title = "$CLOSE_ALL_TABLES_OR_TABS_TITLE" if show_all_services else "$SELECT_TAB"
            index = show_listbox(
                    self.pos_id,
                    listbox_options,
                    message=title
            )
            if index is None:
                raise StopAction()

            return related_objects[index]["table_id"]

        return self.table_id

    def do_total_table(
        self,
        tip_rate,           # type: str
        selected_to_total   # type: str

    ):
        # type: (...) -> str

        return self.table_service.do_total_table(
            pos_id=self.pos_id,
            table_id=self.table_id,
            tip_rate=tip_rate,
            selected_to_total=selected_to_total
        )

    def select_table_or_related_tab_by_dialog(
        self,
        pos_id,     # type: Any[Union[str, int]]
        table_id    # type: Any[Union[str, int]]
    ):
        # type: (...) -> Any[Union[str, int]]

        related_services, _ = self.get_related_services()
        selected_table_id = self.get_selected_related_services(related_services)
        if selected_table_id != table_id:
            result = self.change_selected_table(pos_id, selected_table_id)
            if result:
                return selected_table_id
            else:
                raise StopAction()

        return selected_table_id

    @staticmethod
    def change_selected_table(
            pos_id,
            selected_table_id
    ):

        from table_actions import (
            deselect_table,
            select_table_or_tab,
        )

        deselect_table(pos_id)
        return select_table_or_tab(pos_id, selected_table_id)



import json

import requests
from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import ErrorRetrievingPendingUpdate
from helper import retry, json_serialize
from requests import Response
from application.util import LoggerUtil, settings, ErrorResponseUtil
from typing import Dict, List

NONE_RESPONSE_ERROR = "Could not obtain the request response content"
logger = LoggerUtil.get_logger_name(__name__)


class MediaUpdaterAPIRepository(object):
    def __init__(self, configs):
        # type: (Configurations) -> None

        self.logger = logger
        self.configs = configs
        self.update_type = UpdateType.media
        self.local_configs = configs.updaters[self.update_type.name]
        self.response_util = ErrorResponseUtil()
        self.error_response = None

    def get_images_permissions(self, hashes_to_get_permission):
        # type: (List[str]) -> Dict

        self.logger.info("Obtaining images permissions...")
        settings.set_update_control("catalog", "$DOWNLOAD_IMAGES_AND_CATEGORY")
        url = self._get_download_image_permission_url()
        headers = self._get_app_json_headers()
        data = "{\"images_hashes\":" + json_serialize(hashes_to_get_permission) + "}"

        self.logger.info("Trying get image permissions for: {}".format(data))

        error_message = "Error getting images permissions: {}"
        response = retry(
            logger=self.logger,
            method_name="get_images_permissions",
            n=3,
            sleep_seconds=5,
            trying_to_do=lambda: requests.request("POST", url=url, headers=headers, data=data, timeout=60),
            is_success=lambda resp: resp.status_code == 200,
            success=lambda resp: resp,
            failed=lambda resp: self.logger.error(error_message.format(json_serialize(resp)))
        )

        if response is None:
            raise ErrorRetrievingPendingUpdate(NONE_RESPONSE_ERROR)

        self.logger.info("Permissions obtained!")
        return json.loads(response.content)

    def get_totem_config(self):
        # type: () -> Dict

        self.logger.info("Obtaining images totem...")
        url = self._get_download_image_totem_url()
        headers = self._get_app_json_headers()

        error_message = "Error getting totem config : {}"
        response = retry(
            logger=self.logger,
            method_name="get_images_totem",
            n=3,
            sleep_seconds=5,
            trying_to_do=lambda: requests.request("GET", url=url, headers=headers, timeout=60),
            is_success=lambda resp: resp.status_code == 200,
            success=lambda resp: resp,
            failed=lambda resp: self.logger.error(error_message.format(json_serialize(resp)))
        )

        if response is None:
            raise ErrorRetrievingPendingUpdate(NONE_RESPONSE_ERROR)

        self.logger.info("Permissions obtained!")
        return json.loads(response.content)

    @staticmethod
    def download_image(download_url):
        # type: (str) -> Response

        return requests.get(download_url, allow_redirects=True)

    def _get_app_json_headers(self):
        # type: () -> Dict[str, str]

        return {"X-Api-Key": self.configs.api_key,
                "Content-type": "application/json"}

    def _get_download_image_permission_url(self):
        # type: () -> str

        end_point_url = self.local_configs.endpoints.get_download_image_permission.format(storeId=self.configs.store_id)
        return self.configs.base_url + end_point_url

    def _get_download_image_totem_url(self):
        # type: () -> str

        end_point_url = self.local_configs.endpoints.get_download_image_totem.format(storeId=self.configs.store_id)
        return self.configs.base_url + end_point_url



import json
import logging
import time

from bustoken import TK_NDISCOUNT_GET_BENEFITS
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
)
from sysactions import send_message

from model.customexception import ErrorGettingBenefits


class NDiscountInteractor(object):
    def __init__(self):
        # type: () -> None
        self.logger = logging.getLogger("PosActions")

    def get_benefits(
        self,           # type: NDiscountInteractor
        pos_id,         # type: int
        only_visible,   # type: bool
    ):
        # type: (...) -> json

        self.logger.info("Getting benefits for POS: {}".format(pos_id))
        retry = 0
        while retry < 3:
            time.sleep(0.5 * retry)
            try:
                data = json.dumps(dict(posId=pos_id, only_visible=only_visible))
                response = send_message(
                    dest_name="nDiscount",
                    token=TK_NDISCOUNT_GET_BENEFITS,
                    format=FM_PARAM,
                    data=data,
                    timeout=30000000,
                )
                if response.token == TK_SYS_ACK:
                    return json.loads(response.data)
            except Exception as ex:
                self.logger.warn("Error getting benefits data: {} - Retry attempt: {}".format(ex.message, retry))

            retry += 1

        self.logger.exception("[get_benefits] benefits retry failed")
        raise ErrorGettingBenefits()

    def test_connection(
        self,   # type: NDiscountInteractor
    ):
        # type: (...) -> bool
        try:
            data = json.dumps(dict())
            response = send_message("nDiscount", TK_NDISCOUNT_GET_BENEFITS, FM_PARAM, data, timeout=30000000)
            if response:
                return True
            return False
        except (Exception,) as _:
            self.logger.exception("Unreachable component nDiscount")
            return False



# -*- coding: utf-8 -*-

from actions import mb_context, logger
from bustoken import TK_CHAT_GET_LAST_MESSAGES
from msgbus import FM_PARAM, TK_SYS_ACK, MBTimeout, MBException
from sysactions import action


@action
def update_chat_messages(_):
    try:
        msg = mb_context.MB_EasySendMessage("ChatController", TK_CHAT_GET_LAST_MESSAGES, FM_PARAM)
        if msg.token == TK_SYS_ACK:
            return msg.data
    except (MBTimeout, MBException) as _:
        logger.exception("Error retrieving last messages from ChatController")



from jmespath.compat import with_str_method


class JMESPathError(ValueError):
    pass


@with_str_method
class ParseError(JMESPathError):
    _ERROR_MESSAGE = 'Invalid jmespath expression'
    def __init__(self, lex_position, token_value, token_type,
                 msg=_ERROR_MESSAGE):
        super(ParseError, self).__init__(lex_position, token_value, token_type)
        self.lex_position = lex_position
        self.token_value = token_value
        self.token_type = token_type.upper()
        self.msg = msg
        # Whatever catches the ParseError can fill in the full expression
        self.expression = None

    def __str__(self):
        # self.lex_position +1 to account for the starting double quote char.
        underline = ' ' * (self.lex_position + 1) + '^'
        return (
            '%s: Parse error at column %s, '
            'token "%s" (%s), for expression:\n"%s"\n%s' % (
                self.msg, self.lex_position, self.token_value, self.token_type,
                self.expression, underline))


@with_str_method
class IncompleteExpressionError(ParseError):
    def set_expression(self, expression):
        self.expression = expression
        self.lex_position = len(expression)
        self.token_type = None
        self.token_value = None

    def __str__(self):
        # self.lex_position +1 to account for the starting double quote char.
        underline = ' ' * (self.lex_position + 1) + '^'
        return (
            'Invalid jmespath expression: Incomplete expression:\n'
            '"%s"\n%s' % (self.expression, underline))


@with_str_method
class LexerError(ParseError):
    def __init__(self, lexer_position, lexer_value, message, expression=None):
        self.lexer_position = lexer_position
        self.lexer_value = lexer_value
        self.message = message
        super(LexerError, self).__init__(lexer_position,
                                         lexer_value,
                                         message)
        # Whatever catches LexerError can set this.
        self.expression = expression

    def __str__(self):
        underline = ' ' * self.lexer_position + '^'
        return 'Bad jmespath expression: %s:\n%s\n%s' % (
            self.message, self.expression, underline)


@with_str_method
class ArityError(ParseError):
    def __init__(self, expected, actual, name):
        self.expected_arity = expected
        self.actual_arity = actual
        self.function_name = name
        self.expression = None

    def __str__(self):
        return ("Expected %s %s for function %s(), "
                "received %s" % (
                    self.expected_arity,
                    self._pluralize('argument', self.expected_arity),
                    self.function_name,
                    self.actual_arity))

    def _pluralize(self, word, count):
        if count == 1:
            return word
        else:
            return word + 's'


@with_str_method
class VariadictArityError(ArityError):
    def __str__(self):
        return ("Expected at least %s %s for function %s(), "
                "received %s" % (
                    self.expected_arity,
                    self._pluralize('argument', self.expected_arity),
                    self.function_name,
                    self.actual_arity))


@with_str_method
class JMESPathTypeError(JMESPathError):
    def __init__(self, function_name, current_value, actual_type,
                 expected_types):
        self.function_name = function_name
        self.current_value = current_value
        self.actual_type = actual_type
        self.expected_types = expected_types

    def __str__(self):
        return ('In function %s(), invalid type for value: %s, '
                'expected one of: %s, received: "%s"' % (
                    self.function_name, self.current_value,
                    self.expected_types, self.actual_type))


class EmptyExpressionError(JMESPathError):
    def __init__(self):
        super(EmptyExpressionError, self).__init__(
            "Invalid JMESPath expression: cannot be empty.")


class UnknownFunctionError(JMESPathError):
    pass



# -*- coding: utf-8 -*-

import re
from typing import Dict
from xml.etree import cElementTree as eTree

MUST_BE_LIST = ['TenderHistory', 'SaleLine', 'StateHistory', 'ServiceOrders', 'ServiceTenders', 'PaymentOrders']


def _parse_dict(
    dictionary,    # type: Dict
):
    # type: (...) -> Dict

    for key in dictionary.keys():
        is_int = re.match(r'^\s*\d+\s*$', dictionary[key])
        is_float = dictionary[key].count('.') == 1 and re.match(r'^\s*(\d*\.\d+)|(\d+\.\d*)\s*$', dictionary[key])
        if is_int:
            dictionary[key] = int(dictionary[key])
        elif is_float:
            dictionary[key] = float(dictionary[key])
    return dictionary


def _transform_xml_to_dict(
    xml_object,    # type: eTree
):
    # type: (...) -> Dict

    all_tags = [ch.tag for ch in xml_object.getchildren()]
    item = _parse_dict(dictionary=xml_object.attrib) if len(xml_object.items()) > 0 else {}

    if len(xml_object.getchildren()) > 0:
        for child in xml_object.getchildren():
            if all_tags.count(child.tag) > 0 and not item.get(child.tag):
                item[child.tag] = []

            if type(item.get(child.tag)) == list:
                child_data = _transform_xml_to_dict(xml_object=child)

                if type(child_data) == list:
                    item[child.tag].extend(child_data)
                else:
                    item[child.tag].append(child_data)
            else:
                item[child.tag] = _transform_xml_to_dict(xml_object=child)
    if xml_object.tag == "Properties":
        return item
    item = item if len(item.keys()) > 1 or len(item.values()) == 0 else item.get(item.keys()[0])
    if 'key' in item:
        item[item.pop('key')] = item.pop('value')
    return item


def _fix_redundancies(
    dict_object,   # type: Dict
):
    # type: (...) -> Dict

    for attribute in dict_object.keys():
        attr = dict_object.get(attribute)
        if attribute in MUST_BE_LIST or not isinstance(attr, list):
            continue
        if _is_empty_dict_list(attr):
            dict_object[attribute] = [] if attribute in MUST_BE_LIST else {}
            continue
        if _has_same_keys(attr):
            new_dict = {}
            for item in attr:
                new_dict.update(item)
            dict_object[attribute] = new_dict
    return dict_object

def _is_empty_dict_list(
    value,  # type: str
):
    # type: (...) -> bool

    return len(value) > 0 and value[0] == {}

def _has_same_keys(
    value,  # type: str
):
    # type: (...) -> bool

    keys = [prop.keys()[0] for prop in value]
    return keys.count(keys[0]) != len(keys) or len(keys) == 1

def convert_xml_to_dict(
    xml_object,    # type: eTree
):
    # type: (...) -> Dict

    dict_object = _transform_xml_to_dict(xml_object=xml_object)
    dict_object = _fix_redundancies(dict_object=dict_object)
    if 'Orders' in dict_object.keys():
        if type(dict_object.get('Orders')) != list:
            dict_object['Orders'] = [dict_object.get('Orders')]
        for i, order in enumerate(dict_object.get('Orders')):
            dict_object['Orders'][i] = _fix_redundancies(dict_object=dict_object['Orders'][i])
    return dict_object



import logging

from bustoken import TK_SLCTRL_OMGR_UPDATEORDERTENDERDETAIL
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
)

from sysactions import (
    get_model,
    get_posot,
)

from .. import mb_context


logger = logging.getLogger("PosActions")


def update_tenders(
    pos_id,             # type: str
    order_id,           # type: str
    tender_id,          # type: str
    new_tender_detail   # type: str
):
    # type: (...) -> bool
    model = get_model(pos_id)
    pos_ot = get_posot(model)
    try:
        timeout = 600000000
        params = ('1' if pos_ot.blkopnotify else '0', pos_id, order_id, new_tender_detail)
        req_buf = '\x00'.join((str(p) for p in params))
        result = mb_context.MB_EasySendMessage(
            'ORDERMGR%s' % pos_id,
            TK_SLCTRL_OMGR_UPDATEORDERTENDERDETAIL,
            FM_PARAM,
            req_buf,
            timeout
        )
        if result.token == TK_SYS_ACK:
            return True

    except Exception as _:
        logger.exception("Fail salve receipt cancel: Tender: {}".format(tender_id))

    return False



from abc import abstractmethod

from messagebus import (
    Message,
    MessageBus,
    DefaultToken,
    DataType,
    InternalErrorMessage,
)
from messageprocessorutil import UuidMessageProcessor
from typing import (
    Any,
    Optional,
    List,
)

from devicemanager.interactor.dto import DeviceStatusDto
from ._DeviceStatusFormatter import DeviceStatusFormatter


class DeviceStatusDtoProcessor(UuidMessageProcessor):

    @abstractmethod
    def parse_data(self, data, event_type):
        raise NotImplementedError()

    @abstractmethod
    def call_business(self, obj):
        raise NotImplementedError()

    def format_response(
        self,           # type: DeviceStatusDtoProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        request,        # type: Any
        result,         # type: List[DeviceStatusDto]
    ):
        # type: (...) -> None

        reply_message = Message(
            token=DefaultToken.TK_SYS_ACK,
            data_type=DataType.json,
            data=DeviceStatusFormatter.format_as_json(devices=result),
        )
        message_bus.reply_message(
            message=message,
            reply_message=reply_message,
        )

    def format_exception(
        self,           # type: DeviceStatusDtoProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        input_obj,      # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=InternalErrorMessage(ex=exception),
        )

    def format_parse_exception(
        self,           # type: DeviceStatusDtoProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        data,           # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        return self.format_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            input_obj=data,
            exception=exception,
        )



# onderwijsnummer.py - functions for handling onderwijsnummers
#
# Copyright (C) 2012, 2013 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Onderwijsnummer (the Dutch student identification number).

The onderwijsnummers (education number) is very similar to the BSN (Dutch
citizen identification number), but is for students without a BSN. It uses a
checksum mechanism similar to the BSN.

More information:

* https://nl.wikipedia.org/wiki/Onderwijsnummer

>>> validate('1012.22.331')
'101222331'
>>> validate('100252333')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('1012.22.3333')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('2112.22.337')  # number must start with 10
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

from stdnum.exceptions import *
from stdnum.nl.bsn import checksum, compact
from stdnum.util import isdigits


__all__ = ['compact', 'validate', 'is_valid']


def validate(number):
    """Check if the number is a valid onderwijsnummer. This checks the length
    and whether the check digit is correct and whether it starts with the
    right sequence."""
    number = compact(number)
    if not isdigits(number) or int(number) <= 0:
        raise InvalidFormat()
    if not number.startswith('10'):
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    if checksum(number) != 5:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid onderwijsnummer."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

from cryptography.hazmat.primitives.ciphers.base import (
    AEADCipherContext, AEADDecryptionContext, AEADEncryptionContext,
    BlockCipherAlgorithm, Cipher, CipherAlgorithm, CipherContext
)


__all__ = [
    "Cipher",
    "CipherAlgorithm",
    "BlockCipherAlgorithm",
    "CipherContext",
    "AEADCipherContext",
    "AEADDecryptionContext",
    "AEADEncryptionContext",
]



from logging import Logger

from flask import Response
from mebuhi.event import WaiterRetriever


class ListenProcessor(object):
    def __init__(self, waiter_retriever, logger):
        # type: (WaiterRetriever, Logger) -> None
        self.waiter_retriever = waiter_retriever
        self.logger = logger

    def get_caller_ip(self, request):
        if not request.headers.getlist("X-Forwarded-For"):
            return request.remote_addr
        return request.headers.getlist("X-Forwarded-For")[0]

    def handle(self, path_parameters, request):
        request_sync_id = request.values['syncId']
        caller_ip = self.get_caller_ip(request)
        self.logger.info("IP: {} Start handling SyncId - {}".format(caller_ip, request_sync_id))
        response_sync_id = None
        try:
            waiter = self.waiter_retriever.get_waiter(request_sync_id)
            if waiter is None:
                self.logger.info("No waiter with SyncId - Returning 500")
                return Response(status=500, headers={"x-out-of-sync": "true"})
            response_sync_id = waiter.get_sync_id()
            events = waiter.get_events()
            response = self.create_response_xml(events)
            headers = {"X-sync-id": response_sync_id}
            return Response(response, status=200, content_type="text/xml", headers=headers)
        except (Exception,) as _:
            self.logger.exception("Error processing listen")
        finally:
            log_message = "IP: {} Finish handling SyncId - {} / {}"
            self.logger.info(log_message.format(caller_ip, request_sync_id, response_sync_id))

    def create_response_xml(self, events):
        event_xml = ""
        for event in events:
            event_xml += event.data
        return "<Events>{}</Events>".format(event_xml)



from typing import List

from dailygoals_services.model import SaleChallenge
from dailygoals_services.model.repository import DailyGoalsRepository


class GetSalesChallengeInteractor:
    def __init__(self, daily_goals_repository):
        # type: (DailyGoalsRepository) -> None

        self.daily_goals_repository = daily_goals_repository

    def get_sales_challenge(self, limit):
        # type: (int) -> List[SaleChallenge]

        return self.daily_goals_repository.get_sales_challenge(limit)



from model import BaseThread
from cfgtools import Group
from datetime import datetime
from time import sleep

from rupture_events import RuptureEventTypes
from rupture_service import RuptureService


class CleanRuptureThread(BaseThread):
    def __init__(
        self,
        configs,            # type: Group
        rupture_service,    # type: RuptureService
    ):
        # type: (...) -> None

        super(CleanRuptureThread, self).__init__()

        self._start = configs.find_value("Rupture.CleanRuptureTimeWindow.StartTime", None)
        self._end = configs.find_value("Rupture.CleanRuptureTimeWindow.EndTime", None)
        self._interval = int(configs.find_value("Rupture.CleanRuptureTimeWindow.Interval", "30"))
        self.rupture_service = rupture_service

    def run(self):
        while self.running:
            sleep_time = self._interval

            try:
                now = datetime.now().time().isoformat()
                business_period = datetime.today().strftime("%Y%m%d")
                should_run = self._start and self._end and self._start <= now <= self._end

                if should_run:
                    self.rupture_service.set_last_clean_event(period=business_period)
                    self.rupture_service.clean_rupture_items(RuptureEventTypes.CleanRuptureThread)
            finally:
                sleep(sleep_time)



class DefaultManagerConfig(object):
    def __init__(self):

        self.enabled = None
        self.name = None
        self.active_frequency = None
        self.window = Window()
        self.apply_window = Window()


class Window(object):
    def __init__(self):
        self.start_time = None
        self.end_time = None



from ._TagLinesCommand import TagLinesCommand


class AddTagLinesCommand(TagLinesCommand):
    def __init__(self, order_id, view, line_ids, tag, command_order):
        super(AddTagLinesCommand, self).__init__(order_id, view, line_ids, tag, command_order)

    def get_hash_extra_value(self):
        # type: () -> str

        return "AddTagLinesCommand{}{}".format(self.line_ids, self.tag)



# vat.py - functions for handling European VAT numbers
# coding: utf-8
#
# Copyright (C) 2012-2021 Arthur de Jong
# Copyright (C) 2015 Lionel Elie Mamane
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""VAT (European Union VAT number).

The European Union VAT number consists of a 2 letter country code (ISO
3166-1, except Greece which uses EL) followed by a number that is
allocated per country.

The exact format of the numbers varies per country and a country-specific
check is performed on the number using the VAT module that is relevant for
that country.

>>> compact('ATU 57194903')
'ATU57194903'
>>> validate('BE697449992')
'BE0697449992'
>>> validate('FR 61 954 506 077')
'FR61954506077'
>>> guess_country('00449544B01')
['nl']
"""

from stdnum.eu import oss
from stdnum.exceptions import *
from stdnum.util import clean, get_cc_module, get_soap_client


MEMBER_STATES = set([
    'at', 'be', 'bg', 'cy', 'cz', 'de', 'dk', 'ee', 'es', 'fi', 'fr', 'gr',
    'hr', 'hu', 'ie', 'it', 'lt', 'lu', 'lv', 'mt', 'nl', 'pl', 'pt', 'ro',
    'se', 'si', 'sk', 'xi',
])
"""The collection of country codes that are queried. Greece is listed with a
country code of gr while for VAT purposes el is used instead. For Northern
Ireland numbers are prefixed with xi of United Kingdom numbers."""

_country_modules = dict()

vies_wsdl = 'https://ec.europa.eu/taxation_customs/vies/checkVatService.wsdl'
"""The WSDL URL of the VAT Information Exchange System (VIES)."""


def _get_cc_module(cc):
    """Get the VAT number module based on the country code."""
    # Greece uses a "wrong" country code
    cc = cc.lower()
    if cc in ('eu', 'im'):
        return oss
    if cc == 'el':
        cc = 'gr'
    if cc not in MEMBER_STATES:
        return
    if cc == 'xi':
        cc = 'gb'
    if cc not in _country_modules:
        _country_modules[cc] = get_cc_module(cc, 'vat')
    return _country_modules[cc]


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, '').upper().strip()
    cc = number[:2]
    module = _get_cc_module(cc)
    if not module:
        raise InvalidComponent()
    number = module.compact(number)
    if not number.startswith(cc):
        number = cc + number
    return number


def validate(number):
    """Check if the number is a valid VAT number. This performs the
    country-specific check for the number."""
    number = clean(number, '').upper().strip()
    cc = number[:2]
    module = _get_cc_module(cc)
    if not module:
        raise InvalidComponent()
    number = module.validate(number)
    if not number.startswith(cc):
        number = cc + number
    return number


def is_valid(number):
    """Check if the number is a valid VAT number. This performs the
    country-specific check for the number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def guess_country(number):
    """Guess the country code based on the number. This checks the number
    against each of the validation routines and returns the list of countries
    for which it is valid. This returns lower case codes and returns gr (not
    el) for Greece."""
    return [cc
            for cc in MEMBER_STATES
            if _get_cc_module(cc).is_valid(number)]


def check_vies(number, timeout=30):  # pragma: no cover (not part of normal test suite)
    """Query the online European Commission VAT Information Exchange System
    (VIES) for validity of the provided number. Note that the service has
    usage limitations (see the VIES website for details). The timeout is in
    seconds. This returns a dict-like object."""
    # this function isn't automatically tested because it would require
    # network access for the tests and unnecessarily load the VIES website
    number = compact(number)
    client = get_soap_client(vies_wsdl, timeout)
    return client.checkVat(number[:2], number[2:])


def check_vies_approx(number, requester, timeout=30):  # pragma: no cover
    """Query the online European Commission VAT Information Exchange System
    (VIES) for validity of the provided number, providing a validity
    certificate as proof. You will need to give your own VAT number for this
    to work. Note that the service has usage limitations (see the VIES
    website for details). The timeout is in seconds. This returns a dict-like
    object."""
    # this function isn't automatically tested because it would require
    # network access for the tests and unnecessarily load the VIES website
    number = compact(number)
    requester = compact(requester)
    client = get_soap_client(vies_wsdl, timeout)
    return client.checkVatApprox(
        countryCode=number[:2], vatNumber=number[2:],
        requesterCountryCode=requester[:2], requesterVatNumber=requester[2:])



# cnpj.py - functions for handling CNPJ numbers
# coding: utf-8
#
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CNPJ (Cadastro Nacional da Pessoa Jurídica, Brazilian company identifier).

Numbers from the national register of legal entities have 14 digits. The
first 8 digits identify the company, the following 4 digits identify a
business unit and the last 2 digits are check digits.

>>> validate('16.727.230/0001-97')
'16727230000197'
>>> validate('16.727.230.0001-98')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('16.727.230/0001=97')  # invalid delimiter
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> format('16727230000197')
'16.727.230/0001-97'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -./').strip()


def calc_check_digits(number):
    """Calculate the check digits for the number."""
    d1 = (11 - sum(((3 - i) % 8 + 2) * int(n)
                   for i, n in enumerate(number[:12]))) % 11 % 10
    d2 = (11 - sum(((4 - i) % 8 + 2) * int(n)
                   for i, n in enumerate(number[:12])) -
          2 * d1) % 11 % 10
    return '%d%d' % (d1, d2)


def validate(number):
    """Check if the number is a valid CNPJ. This checks the length and
    whether the check digits are correct."""
    number = compact(number)
    if not isdigits(number) or int(number) <= 0:
        raise InvalidFormat()
    if len(number) != 14:
        raise InvalidLength()
    if calc_check_digits(number) != number[-2:]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid CNPJ."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return (number[0:2] + '.' + number[2:5] + '.' + number[5:8] + '/' +
            number[8:12] + '-' + number[12:])



from production.box import ProductionBox
from typing import Union, List


class ProductionTree(object):
    def __init__(self, name, root, leaves):
        # type: (str, ProductionBox, Union[ProductionBox, List[ProductionBox]]) -> None
        self.name = name
        self.root = root
        self.leaves = leaves
        self.tree_name = ''

    def set_tree_name(self, tree_name):
        # type: (str) -> None

        self.tree_name = tree_name



# mac.py - functions for handling MAC (Ethernet) addresses
#
# Copyright (C) 2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""MAC address (Media Access Control address).

A media access control address (MAC address, sometimes Ethernet address) of a
device is meant as a unique identifier within a network at the data link
layer.

More information:

* https://en.wikipedia.org/wiki/MAC_address
* https://en.wikipedia.org/wiki/Organizationally_unique_identifier
* https://standards.ieee.org/faqs/regauth.html#2

>>> validate('D0-50-99-84-A2-A0')
'd0:50:99:84:a2:a0'
>>> to_eui48('d0:50:99:84:a2:a0')
'D0-50-99-84-A2-A0'
>>> is_multicast('d0:50:99:84:a2:a0')
False
>>> str(get_manufacturer('d0:50:99:84:a2:a0'))
'ASRock Incorporation'
>>> get_oui('d0:50:99:84:a2:a0')
'D05099'
>>> get_iab('d0:50:99:84:a2:a0')
'84A2A0'
"""

import re

from stdnum import numdb
from stdnum.exceptions import *
from stdnum.util import clean


_mac_re = re.compile('^([0-9a-f]{2}:){5}[0-9a-f]{2}$')


def compact(number):
    """Convert the MAC address to the minimal, consistent representation."""
    number = clean(number, ' ').strip().lower().replace('-', ':')
    # zero-pad single-digit elements
    return ':'.join('0' + n if len(n) == 1 else n for n in number.split(':'))


def _lookup(number):
    """Look up the manufacturer in the IEEE OUI registry."""
    number = compact(number).replace(':', '').upper()
    info = numdb.get('oui').info(number)
    try:
        return (
            ''.join(n[0] for n in info[:-1]),
            info[-2][1]['o'].replace('%', '"'))
    except IndexError:
        raise InvalidComponent()


def get_manufacturer(number):
    """Look up the manufacturer in the IEEE OUI registry."""
    return _lookup(number)[1]


def get_oui(number):
    """Return the OUI (organization unique ID) part of the address."""
    return _lookup(number)[0]


def get_iab(number):
    """Return the IAB (individual address block) part of the address."""
    number = compact(number).replace(':', '').upper()
    return number[len(get_oui(number)):]


def is_unicast(number):
    """Check whether the number is a unicast address.

    Unicast addresses are received by one node in a network (LAN)."""
    number = compact(number)
    return int(number[:2], 16) & 1 == 0


def is_multicast(number):
    """Check whether the number is a multicast address.

    Multicast addresses are meant to be received by (potentially) multiple
    nodes in a network (LAN)."""
    return not is_unicast(number)


def is_broadcast(number):
    """Check whether the number is the broadcast address.

    Broadcast addresses are meant to be received by all nodes in a network."""
    number = compact(number)
    return number == 'ff:ff:ff:ff:ff:ff'


def is_universally_administered(number):
    """Check if the address is supposed to be assigned by the manufacturer."""
    number = compact(number)
    return int(number[:2], 16) & 2 == 0


def is_locally_administered(number):
    """Check if the address is meant to be configured by an administrator."""
    return not is_universally_administered(number)


def validate(number, validate_manufacturer=None):
    """Check if the number provided is a valid MAC address.

    The existence of the manufacturer is by default only checked for
    universally administered addresses but can be explicitly set with the
    `validate_manufacturer` argument.
    """
    number = compact(number)
    if len(number) != 17:
        raise InvalidLength()
    if not _mac_re.match(number):
        raise InvalidFormat()
    if validate_manufacturer is not False:
        if validate_manufacturer or is_universally_administered(number):
            get_manufacturer(number)
    return number


def is_valid(number, validate_manufacturer=None):
    """Check if the number provided is a valid IBAN."""
    try:
        return bool(validate(number, validate_manufacturer=validate_manufacturer))
    except ValidationError:
        return False


def to_eui48(number):
    """Convert the MAC address to EUI-48 format."""
    return compact(number).upper().replace(':', '-')



import base64
import cPickle as Pickle
import os
from Queue import Queue, Empty
from datetime import (
    datetime,
    timedelta,
)
from threading import (
    Thread,
    Lock,
)

import production.repository
from mw_helper import BaseRepository
from persistence import Connection
from production.model import (
    Item,
    ProductionOrder,
)
from typing import (
    List,
    Dict,
    Optional,
)


lock_db = Lock()


class ProductionRepository(BaseRepository, production.repository.ProductionRepository):
    def __init__(self, mb_context, logger):
        super(ProductionRepository, self).__init__(mb_context)
        self.logger = logger

        self.order_queue = Queue()
        self.running = True

        self.processor_thread = Thread(target=self.consume_queue)
        self.processor_thread.daemon = True
        self.processor_thread.start()

    def consume_queue(self):
        while self.running:
            order_id = None
            try:
                order_to_save = self.order_queue.get(True, 5)
                order_id = order_to_save.order_id
                self.save_order_async(order_to_save)
            except Empty:
                continue
            except BaseException as _:  # noqa
                self.logger.exception("Error saving order: {}".format(order_id))

    def stop(self):
        self.running = False

    def save_order(
        self,       # type: ProductionRepository
        order,      # type: ProductionOrder
    ):
        # type: (...) -> None

        try:
            self.remove_forbidden_tags(
                items=order.items,
                order_id=str(order.order_id),
            )
        except Exception as _:
            self.logger.error("Error handling remove_forbidden_tags for order id {}".format(order.order_id))

        self.order_queue.put(order)

    def remove_forbidden_tags(
        self,           # type: ProductionRepository
        items,          # type: List[Item]
        order_id,       # type: str
    ):
        # type: (...) -> None

        forbidden_tags = [
            "wait-prep-time"
        ]

        for item in items:
            for forbidden_tag in forbidden_tags:
                if item.has_tag(forbidden_tag):
                    item.remove_tag(forbidden_tag)
                    self.logger.error("Denied TAG for Order ID {}: {}".format(order_id, forbidden_tag))

            if len(item.items) > 0:
                self.remove_forbidden_tags(
                    items=item.items,
                    order_id=order_id,
                )

    def save_order_async(self, order):
        def func(conn):
            # type: (Connection) -> None
            order_data = base64.b64encode(Pickle.dumps(order, Pickle.HIGHEST_PROTOCOL))
            conn.query(insert_query.format(order.order_id, order_data, order.created_at, 0))
            self.logger.info("Saving order: {} to Database".format(order.order_id))

        self.execute_with_connection(func)

    def get_all_orders(self):
        def func(conn):
            # type: (Connection) -> List[ProductionOrder]
            orders = []
            for row in conn.select(select_query):
                order_data = base64.b64decode(row.get_entry(0))
                order = Pickle.loads(order_data)
                orders.append(order)

            return orders

        return self.execute_with_connection(func)

    def get_orders_to_purge(self, limit_date):
        def func(conn):
            # type: (Connection) -> List[int]
            orders = []
            formatted_limit_date = limit_date.strftime("%Y-%m-%dT%H:%M:%S")
            for row in conn.select(get_orders_to_purge_query.format(formatted_limit_date)):
                order_id = int(row.get_entry("OrderId"))
                orders.append(order_id)

            return orders

        return self.execute_with_connection(func)

    def purge_orders(
        self,    # type: ProductionRepository
        orders,  # type: List
    ):
        # type: (...) -> None

        with lock_db:
            def func(
                conn,  # type: Connection
            ):
                # type: (...) -> None

                orders_list = [str(order) for order in orders]
                query = purge_orders_query.format(",".join(orders_list))
                conn.query(sql=query)

            return self.execute_with_connection(method=func)

    def save_pick_list_order(
        self,
        order_id,           # type: int
        order_id_count,     # type: str
        jit,                # type: str
        course,             # type: str
        receipt,            # type: str
        status,             # type: bool
    ):
        # type: (...) -> None

        with lock_db:
            def func(
                conn,   # type: Connection
            ):
                # type: (...) -> None

                date = datetime.now().strftime(
                    "%m/%d/%Y %H:%M:%S",
                )

                query = insert_pick_list_query.format(
                    order_id,
                    order_id_count,
                    jit,
                    course,
                    receipt,
                    date,
                    status
                )

                conn.query(
                    sql=query,
                )

            self.execute_with_connection(func)

    def get_picklist_id(
        self,
        order_id,           # type: int
        order_id_count,     # type: str
        jit,                # type: str
        course,             # type: str
    ):
        # type: (...) -> Optional[int]

        def func(conn):
            # type: (Connection) -> Optional[int]

            query = select_get_pick_list_id_query.format(order_id, order_id_count, jit, course)

            for row in conn.select(query):
                picklist_id = row.get_entry(col="Id")
                return int(picklist_id)

            return None

        return self.execute_with_connection(method=func)

    def get_all_pick_list_orders(
        self,
        orders_list,    # type: List[str]
    ):
        # type: (...) -> Dict

        def func(conn):
            # type: (Connection) -> Dict

            orders = [str(order) for order in orders_list]

            query = select_pick_list_orders_query.format(",".join(orders))

            pick_list_orders = {}

            for row in conn.select(query):
                jit = row.get_entry(
                     col="JIT",
                )

                course_name = row.get_entry(
                   col="Course",
                )
                if course_name != '':
                    course_name = " - {}".format(
                        course_name,
                    )

                key = "{}{}".format(
                    jit,
                    course_name,
                )

                pick_list_orders[key] = row.get_entry(
                        col="Id",
                    )

            return pick_list_orders

        return self.execute_with_connection(func)

    def get_data_pick_list_order(
        self,
        pick_list_order_id,     # type: int
    ):
        # type: (...) -> str

        def func(conn):
            # type: (Connection) -> Optional[str]

            query = """SELECT Data FROM PicklistOrders WHERE Id == {}"""

            query = query.format(
                pick_list_order_id,
            )

            for row in conn.select(query):
                return row.get_entry(
                     col="Data",
                )

            return None

        return self.execute_with_connection(func)

    def delete_outdated_pick_list_orders(
        self,
        range_days,  # type: int
    ):
        # type: (...) -> None

        with lock_db:
            self.logger.info("Start delete_outdated_pick_list_orders")

            def inner_func(
                conn,  # type: Connection
            ):
                # type: (...) -> None

                result_datetime = datetime.utcnow() - timedelta(days=range_days)
                formatted_date = result_datetime.strftime('%m/%d/%y %H:%M:%S')
                query = purge_picklist_orders_query.format(formatted_date)
                conn.query(sql=query)

            self.execute_with_transaction(method=inner_func)
            self.logger.info("End delete_outdated_pick_list_orders")

    def vacuum_database(
        self,
    ):
        # type: (...) -> Dict

        with lock_db:
            db_info = {}

            def inner_func(
                conn,  # type: Connection
            ):
                # type: (...) -> None

                db_info["db_file"] = self._get_db_file(
                    conn=conn,
                    db_name_to_find="production",
                )
                db_info["old_file_size"] = self._get_file_size(file_name=db_info.get("db_file"))

                query = """
                    VACUUM production;
                """
                conn.query(sql=query)

            self.execute_with_connection(method=inner_func)
            db_info["new_file_size"] = self._get_file_size(file_name=db_info.get("db_file"))
            self.logger.info("Finished [vacuum_production_database]")

            return db_info

    @staticmethod
    def _get_db_file(
        conn,
        db_name_to_find,
    ):

        db_file = None
        pragma_cursor = conn.select("PRAGMA database_list;")
        for pragma_row in pragma_cursor:
            db_name, db_file = map(pragma_row.get_entry, (1, 2,))
            if db_name == db_name_to_find:
                break
        return db_file

    @staticmethod
    def _get_file_size(
        file_name,  # type: str
    ):
        # type: (...) -> int

        return os.path.getsize(file_name)


insert_query = """
INSERT OR REPLACE INTO ProductionOrders
(OrderId, Object, CreatedAt, Purged)
VALUES
({}, '{}', '{}', {})
"""

select_query = """
SELECT Object FROM ProductionOrders WHERE PURGED = 0 ORDER BY CreatedAt DESC LIMIT 1000
"""

get_orders_to_purge_query = """
SELECT OrderId FROM ProductionOrders WHERE CreatedAt < '{}' ORDER BY CreatedAt ASC LIMIT 1000
"""

purge_orders_query = """
DELETE FROM ProductionOrders WHERE OrderId IN ({})
"""

insert_pick_list_query = """
INSERT OR REPLACE INTO PicklistOrders
(OrderId, OrderIdCount, JIT, Course, Data, CreatedAt, PrintOK)
VALUES
({}, '{}', '{}', '{}', '{}', '{}', {})
"""

select_get_pick_list_id_query = """
SELECT Id FROM PicklistOrders WHERE OrderId = '{0}' AND OrderIdCount = '{1}' AND JIT = '{2}' AND Course = '{3}'
"""

purge_picklist_orders_query = """
DELETE FROM PicklistOrders WHERE CreatedAt <= '{}'
"""

select_pick_list_orders_query = """
SELECT Id, JIT, Course FROM PicklistOrders WHERE OrderId in ({}) ORDER BY CreatedAt
"""




from production.box import PodTypeFilterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class PodTypeFilterBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(PodTypeFilterBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> PodTypeFilterBox
        return PodTypeFilterBox(config.name,
                                config.sons,
                                self.get_extra(config, "AllowedPodTypes", []),
                                self.get_extra(config, "ExcludedPodTypes", []),
                                self.get_logger(config))



import logging
import sysactions

from actions.util import get_authorization
from mw_helper import show_filtered_list_box_dialog

from typing import (
    Dict,
    List,
    Optional,
    Union,
)
from xml.etree import cElementTree as eTree

logger = logging.getLogger("PosActions")


def get_tender_type(
    pos_id,             # type: str
    model,              # type: eTree.Element
    tender_type_id,     # type: str
):
    # type: (...) -> Optional[Dict[str, Union[str, int, bool]]]

    tender_type = _select_tender_type(
        pos_id=pos_id,
        tender_type_id=tender_type_id,
    )
    if tender_type is None:
        return None

    necessary_level = int(tender_type['needAuth'] or 0)
    operator_level = _get_current_operator_level(model=model)
    if necessary_level > operator_level:
        authorized = get_authorization(
            pos_id=pos_id,
            min_level=necessary_level,
            model=model,
        )[0]

        if not authorized:
            return None

    return tender_type


def get_all_tender_types():
    # type: (...) -> List[Dict[str, Union[str, int, bool]]]

    all_tender_types = sysactions.get_tender_types()
    tender_types = [tender for tender_id, tender in all_tender_types.items() if tender_id < 200]
    for tender_type in tender_types:
        tender_type["showInFC"] = True if tender_type["id"] < 90 else False
        tender_type["showInDL"] = True if 90 <= tender_type["id"] < 120 else False
        parent_id = "parentId "
        if parent_id in tender_type:
            tender_type["parentId"] = tender_type[parent_id]
            del tender_type[parent_id]

        has_child = False
        for tender_type_aux in tender_types:
            if "parentId" in tender_type_aux and tender_type_aux["parentId"] == tender_type["id"]:
                has_child = True
                break

        tender_type["hasChild"] = has_child

    try:
        tender_types = sorted(tender_types, key=lambda k: k['id'])
    except (Exception, ) as _:
        logger.exception("Error ordering tender types")

    return tender_types


def _select_tender_type(
    pos_id,             # type: str
    tender_type_id,     # type: str
):
    # type: (...) -> Optional[Dict[str, Union[str, int, bool]]]

    all_tender_types = get_all_tender_types()
    tender_types = [
        tender
        for tender in all_tender_types
        if (tender["parentId"] is not None and tender["parentId"] == int(tender_type_id))
    ]
    if len(tender_types) != 0:
        if len(tender_types) == 1:
            return tender_types[0]
        else:
            options = [tender["descr"] for tender in tender_types]
            selected_option = show_filtered_list_box_dialog(
                pos_id=pos_id,
                options=options,
                title="$SELECT_PAYMENT_TYPE",
                mask="NOFILTER",
            )
            if selected_option is None:
                return None

            tender_type_id = tender_types[selected_option]["id"]
    else:
        return [tender for tender in all_tender_types if tender["id"] == int(tender_type_id)][0]

    return [tender for tender in tender_types if tender["id"] == tender_type_id][0]


def _get_current_operator_level(
    model,  # type: eTree.Element
):
    # type: (...) -> int

    operator_level = 0
    model_operator = model.find("Operator")
    if model_operator is not None:
        operator_level = int(model_operator.get("level") or 0)

    return operator_level



# -*- coding: utf-8 -*-

import logging

from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, DataType
from msgbus import TK_STORECFG_GET, TK_SYS_NAK
from poshelper import config_logger_with_hv_logger
from typing import (
    Optional,
    Dict,
)

from application.model.customexception import ConfigurationReaderException
from application.model.configuration import (
    Configurations,
    EndpointConfig,
)
from cfgtools import read

endpoints_config_name = ["AskTotaled", "AskPayment"]


class Configurator(object):

    def __init__(self, loader_path):
        # type: (str) -> None

        self.loader_path = loader_path
        self.reader = read(self.loader_path)
        self.configs = Configurations()

    def read_loader_configurations(self):
        # type: () -> None

        try:
            self.configs.service_name = str(self.reader.find_value("Identification.ComponentName"))
            self.configs.persistence_name = str(self._get_config("PersistenceName") or "")
            self.configs.required_services = "|".join(self._get_config("RequiredServices", array=True) or [])

            self._set_credentials_by_types()
            self.configs.endpoints.retrieve_loyalty_customer_info = str(self._get_config("EndPoints.RetrieveLoyaltyId") or "")
            self.configs.endpoints.get_benefit = str(self._get_config("EndPoints.GetBenefit") or "")
            self.configs.endpoints.burn_benefit = str(self._get_config("EndPoints.BurnBenefit") or "")
            self.configs.endpoints.lock_benefit = str(self._get_config("EndPoints.LockBenefit") or "")
            self.configs.endpoints.unlock_benefit = str(self._get_config("EndPoints.ReleaseBenefit") or "")
            self.configs.endpoints.get_totaled_message = str(self._get_config("EndPoints.Totaled") or "")
            self.configs.endpoints.payment = str(self._get_config("EndPoints.Payment") or "")

            self.configs.is_loyalty_enabled = str(self._get_config("LoyaltyEnabled") or "true").lower() == "true"
            if self.configs.is_loyalty_enabled:
                self.configs.id_type = str(self._get_config("IdentifierType") or "DOCUMENT")

            self.configs.endpoints_config = self._parse_endpoints_config()

        except Exception as ex:
            raise ConfigurationReaderException(ex)

    def _parse_endpoints_config(
        self,    # type: Configurator
    ):
        # type: (...) -> Dict[str, EndpointConfig]

        endpoints_config_group = self.reader.find_group("Loyalty.EndPointsConfigurations")
        dict_configs = {}

        if endpoints_config_group:
            for endpoint_config in endpoints_config_group.keys:
                dict_configs[endpoint_config.name] = EndpointConfig(
                    name=endpoint_config.name,
                    call_external_endpoint=endpoint_config.values[0].lower() == "true",
                )
        else:
            for endpoint_config_name in endpoints_config_name:
                dict_configs[endpoint_config_name] = EndpointConfig(name=endpoint_config_name)

        return dict_configs

    def configure_logger(self):
        # type: () -> None

        config_logger_with_hv_logger(
            loader_path=self.loader_path,
            log_name=self.configs.service_name,
            configure_root=True,
            root_level=logging.DEBUG,
        )
        self.configs.logger = logging.getLogger(self.configs.service_name)

    def get_store_id(self, message_bus):
        # type: (MbContextMessageBus) -> None

        self.configs.store_id = self._read_sw_config(message_bus, "Store.Id")

    @staticmethod
    def _read_sw_config(message_bus, key):
        # type: (MbContextMessageBus, str) -> Configurations

        message = Message(TK_STORECFG_GET, data_type=DataType.param, data=key)
        msg = message_bus.send_message("StoreWideConfig", message)
        if msg.token == TK_SYS_NAK:
            raise ConfigurationReaderException("Unable to get Store Wide Configuration: {}".format(key))
        return str(msg.data)

    def _get_config(
        self,               # type: Configurator
        key,                # type: str
        group_name="",      # type: Optional[str]
        array=False,        # type: Optional[bool]
    ):
        # type: (...) -> str

        group_replace = "." if not group_name else ".{}.".format(group_name)
        path = self.configs.service_name + group_replace + key
        return self.reader.find_values(path) if array else self.reader.find_value(path)

    def _set_credentials_by_types(
        self,   # type: Configurator
    ):
        # type: (...) -> None

        for applier_sub_type, credential in self.configs.credentials.items():
            config_value = self._get_config(
                key=applier_sub_type + ".BaseURL",
                group_name="EndPointsCredentials",
            )
            credential.base_url = str(config_value or "").strip("/")
            config_value = self._get_config(
                key=applier_sub_type + ".ApiKey",
                group_name="EndPointsCredentials",
            )
            credential.api_key = str(config_value or "")



import persistence

from typing import Any

from sysactions import action
from application.interactor import FiscalWrapperInteractor
from .. import logger, mb_context


nf_type = None


@action
def get_nf_type(*args):
    # type: (Any) -> str

    global nf_type
    if not nf_type:
        fiscal_wrapper_interactor = FiscalWrapperInteractor()
        nf_type = fiscal_wrapper_interactor.get_nf_type()
    return nf_type


def populate_combo_taxes():
    BEGIN_TRANSACTION = ["BEGIN TRANSACTION", "UPDATE fiscalinfo.FiscalDEnabled SET Enabled=0"]
    COMMIT_TRANSACTION = ["UPDATE fiscalinfo.FiscalDEnabled SET Enabled=1", "COMMIT TRANSACTION"]

    query = """
        INSERT OR REPLACE INTO taxcalc.ProductTaxCategory(ItemId, TaxCgyId)
        SELECT PI.ItemId, V1.TaxCgyId
            FROM (
                SELECT P.ItemId, P.ProductCode, PTC.TaxCgyId
                FROM sysproddb.ProductItem P
                LEFT JOIN taxcalc.ProductTaxCategory PTC ON PTC.ItemId=P.ItemId
                WHERE P.RecursionLevel=1 AND PTC.TaxCgyId IS NOT NULL
            ) V1
        JOIN sysproddb.ProductItem PI ON PI.ProductCode=V1.ProductCode
        WHERE PI.RecursionLevel>1;
        INSERT OR REPLACE INTO fiscalinfo.FiscalProductCatalogChangesCounter SELECT * FROM productdb.ProductCatalogChangesCounter;

        DELETE FROM fiscalinfo.FiscalProductDB
        WHERE EXISTS (
            SELECT 1 FROM productdb.ProductCatalogChangesCounter A
            LEFT JOIN fiscalinfo.FiscalProductCatalogChangesCounter B
            WHERE A.ChangesCount!=COALESCE(B.ChangesCount,-1) OR A.ChangesDateTime!=COALESCE(B.ChangesDateTime,''));

        INSERT OR REPLACE INTO fiscalinfo.FiscalProductDB(ProductCode,ProductName,UnitPrice,TaxFiscalIndex,FiscalCategory,TaxRate,IAT,IPPT,MeasureUnit)
        SELECT ProductCode,ProductName,UnitPrice,TaxFiscalIndex,FiscalCategory,TaxRate,IAT,IPPT,MeasureUnit
        FROM
            (SELECT
                    Price.ProductCode AS ProductCode,
                    Product.ProductName AS ProductName,
                    Price.DefaultUnitPrice as UnitPrice,
                    TaxRule.TaxFiscalIndex AS TaxFiscalIndex,
                    TaxCategory.TaxCgyDescr AS FiscalCategory,
                    tdscale(TaxRule.TaxRate,2,0) AS TaxRate,
                    COALESCE(IAT.CustomParamValue,'T') AS IAT,
                    COALESCE(IPPT.CustomParamValue,'P') AS IPPT,
                    COALESCE(PKP.MeasureUnit,'UN') AS MeasureUnit
                FROM
                    Price Price
                LEFT JOIN taxcalc.ProductTaxCategory ProductTaxCategory
                    ON ProductTaxCategory.ItemId=(Price.Context || "." || Price.ProductCode)
                LEFT JOIN taxcalc.TaxCategory TaxCategory
                    ON TaxCategory.TaxCgyId=ProductTaxCategory.TaxCgyId
                LEFT JOIN taxcalc.TaxRule TaxRule
                    ON TaxRule.TaxCgyId=ProductTaxCategory.TaxCgyId
                LEFT JOIN Product
                    ON Price.ProductCode=Product.ProductCode
                LEFT JOIN productdb.ProductKernelParams PKP
                    ON PKP.ProductCode=Product.ProductCode
                LEFT JOIN productdb.ProductCustomParams IAT
                    ON IAT.ProductCode=Product.ProductCode AND IAT.CustomParamId='IAT'
                LEFT JOIN productdb.ProductCustomParams IPPT
                    ON IPPT.ProductCode=Product.ProductCode AND IPPT.CustomParamId='IPPT'
                WHERE CURRENT_TIMESTAMP BETWEEN Price.ValidFrom AND Price.ValidThru
            ) FiscalProductsView
        WHERE EXISTS (
            SELECT 1 FROM productdb.ProductCatalogChangesCounter A
            LEFT JOIN fiscalinfo.FiscalProductCatalogChangesCounter B
            WHERE A.ChangesCount!=COALESCE(B.ChangesCount,-1) OR A.ChangesDateTime!=COALESCE(B.ChangesDateTime,'')
        );

        INSERT INTO fiscalinfo.FiscalProductCatalogChangesCounter SELECT * FROM productdb.ProductCatalogChangesCounter;
    """

    queries = [query]
    conn = None
    try:
        conn = persistence.Driver().open(mb_context)
        queries = BEGIN_TRANSACTION + queries + COMMIT_TRANSACTION
        conn.query("\0".join(queries))
    except Exception:
        logger.exception("ERRO CRIANDO FISCALINFO DATABASE INFORMATION")
    finally:
        if conn:
            conn.close()



from . import StatisticBox


def format_seconds(seconds):
    # type: (float) -> str
    only_seconds = int(seconds % 60)
    only_minutes = int(seconds / 60)
    return '{:02d}:{:02d}'.format(only_minutes, only_seconds)


class MeanTimeStatisticBox(StatisticBox):
    def __init__(self, name, sons, statistic_updater, logger, get_config):
        super(MeanTimeStatisticBox, self).__init__(name, sons, statistic_updater, logger, get_config)
        self.statistics = {}
        self.tag = get_config("TagName")

    def update_statistics(self, order):
        order_id = order.order_id

        self.debug('Got order! Order: {}', order)

        if self.tag not in order.tags:
            if order_id in self.statistics:
                del self.statistics[order_id]
        else:
            self.statistics[order_id] = order.tags[self.tag]
            self.debug('added order to statistics')
            service_mean_time = sum(self.statistics.itervalues()) / (len(self.statistics) or 1)
            return format_seconds(service_mean_time)



from pos_model import Order
from typing import (
    Any,
    Dict,
)

from _ContextKeys import ContextKeys
from _NfceXmlPartBuilder import NfceXmlPartBuilder
from services import FeeService


class DetFeeBuilder(NfceXmlPartBuilder):

    def __init__(
        self,           # type: DetFeeBuilder
        det_builder,    # type: NfceXmlPartBuilder
        fee_service,    # type: FeeService
    ):
        # type: (...) -> None

        self.det_builder = det_builder
        self.fee_service = fee_service

    def build_xml(
        self,       # type: DetFeeBuilder
        order,      # type: Order
        context,    # type: Dict[unicode, Any]
    ):
        # type: (...) -> unicode

        xml = ""
        fee_items = self.fee_service.get_fee_sale_items(order=order)

        for item in fee_items:
            context[ContextKeys.current_sale_item] = item
            xml += self.det_builder.build_xml(
                order=order,
                context=context,
            )
            context[ContextKeys.item_no] += 1
        return xml



from typing import Any

from application.model import AsyncUuidMessageProcessor
from messagebus import (
    DefaultToken,
    DataType,
    Message,
    MessageBus,
    Event
)
from application.interactor import ChromeUpdaterInteractor


class ChromeUpdaterProcessor(AsyncUuidMessageProcessor):
    def __init__(
        self,
        interactor    # type: ChromeUpdaterInteractor
    ):
        # type: (...) -> None

        super(ChromeUpdaterProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(
        self,
        data,         # type: Any
        event_type    # type: Any
    ):
        # type: (...) -> bool

        return str(data) != "noValidation"

    def call_business(
        self,
        data     # type: str
    ):
        # type: (...) -> Any

        return self.interactor.update_chrome(data)

    @staticmethod
    def format_response(
        message_bus,  # type: MessageBus
        message,      # type: Message
        event,        # type: Event
        _,
        response      # type: Any
    ):
        # type: (...) -> None

        reply_message = Message(
            token=DefaultToken.TK_SYS_ACK,
            data_type=DataType.param,
            data=response
        )
        if message:
            message_bus.reply_message(
                message=message,
                reply_message=reply_message
            )
        elif event:
            message_bus.reply_event(
                event=event,
                reply_message=reply_message
            )

    def format_exception(
        self,
        message_bus,      # type: MessageBus
        message,          # type: Message
        event,            # type: Event
        _,
        exception         # type: Any
    ):
        # type: (...) -> None

        self._reply_exception(
            event=event,
            exception=exception,
            message=message,
            message_bus=message_bus
        )

    def format_parse_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        _,
        exception       # type: Any
    ):
        # type: (...) -> None

        self._reply_exception(
            event=event,
            exception=exception,
            message=message,
            message_bus=message_bus
        )

    @staticmethod
    def _reply_exception(
        event,          # type: Event
        exception,      # type: Any
        message,        # type: Message
        message_bus     # type: MessageBus
    ):
        # type: (...) -> None

        reply_message = Message(
            token=DefaultToken.TK_SYS_NAK,
            data_type=DataType.param,
            data=str(exception)
        )
        if message:
            reply_message = Message(
                token=DefaultToken.TK_SYS_NAK,
                data_type=DataType.param,
                data=str(type(exception).__name__)
            )
            message_bus.reply_message(
                message=message,
                reply_message=reply_message
            )
        elif event:
            message_bus.reply_event(
                event=event,
                reply_message=reply_message
            )

    @staticmethod
    def _do_not_remove_this_function_bug_in_the_library():
        return ''



import os

import cfgtools

from datetime import (
    date,
    datetime,
)
from typing import List

from actions.rupture.get_ruptured_products import has_rupture_items
from actions.rupture.clean_rupture import clean_rupture_on_open_day

from actions.update import (
    get_catalog_version_to_apply,
    get_message_update,
    apply_update,
)
from helper import translate_pos_label
from msgbus import (
    TK_SYS_ACK,
    MBEasyContext,
)
from sysactions import (
    show_messagebox,
    format_date,
    StopAction,
    get_model,
    get_poslist,
    get_business_period,
    get_podtype,
)

BUTTONS = "$OK|$CANCEL"
total_blocking_days_cache = None


class OpenDayInteractor(object):
    def __init__(
        self,
        pos_list=None,      # type: List
    ):
        # type: (...) -> None

        self.pos_list = pos_list
        self.current_pos_id = None
        self.qa_test_runner = os.environ.get("QA_TEST_RUNNER", False)

        if not self.pos_list:
            self.pos_list = get_poslist()

    def process_before_open_day(
        self,                # type: OpenDayInteractor
        pos_id,              # type: int
        mb_context,          # type: MBEasyContext
    ):
        # type: (...) -> None

        self.current_pos_id = pos_id

        self._check_has_some_opened_day_blocking_opening_business_day()
        self._check_has_menu_update(pos_id=pos_id)
        self._confirm_open_day(pos_id=pos_id)

        if self._is_first_pos_opening_current_day():
            clean_rupture_on_open_day()
            self.send_open_day_event(mb_context=mb_context)

    def _check_has_menu_update(
        self,    # type: OpenDayInteractor
        pos_id,  # type: int
    ):
        # type: (...) -> None

        if self.qa_test_runner:
            return

        msg = get_catalog_version_to_apply()
        if msg.token != TK_SYS_ACK:
            return

        get_message_update(pos_id, "$APPLYING_UPDATE_MENU_VERSION_PENDING|" + msg.data)
        has_rupture_items(pos_id)
        applied_update = apply_update(pos_id)
        if applied_update:
            return

        confirm = show_messagebox(pos_id, message="$UPDATE_MENU_ERROR_PROCEED", buttons=BUTTONS)
        if (confirm is None) or (confirm == 1):
            raise StopAction()

    @staticmethod
    def _confirm_open_day(
        pos_id,  # type: int
    ):
        # type: (...) -> None

        today = date.today().strftime("%Y%m%d")
        confirm_message = "$CONFIRM_OPEN_CURRENT_DAY|{0}".format(format_date(today))
        confirm = show_messagebox(pos_id, message=confirm_message, buttons=BUTTONS)
        if (confirm is None) or (confirm == 1):
            raise StopAction()

    @staticmethod
    def _is_first_pos_opening_current_day():
        # type: (...) -> bool

        pos_list = sorted(get_poslist())
        current_business_period = date.today().strftime("%Y%m%d")

        for pos in pos_list:
            model = get_model(posid=pos)
            business_period = get_business_period(model=model)
            if current_business_period == business_period:
                return False

        return True

    @staticmethod
    def send_open_day_event(
        mb_context,     # type: MBEasyContext
    ):
        # type: (...) -> None

        current_business_period = date.today().strftime("%Y%m%d")
        event_xml = '''<STORE_OPEN businessPeriod="{}"/>'''
        data = event_xml.format(current_business_period)

        mb_context.MB_EasyEvtSend(
            subject="STORE_STATUS",
            type="OPEN_STORE_DAY",
            xml=data,
            synchronous=False,
            timeout=-1,
        )

    def _check_has_some_opened_day_blocking_opening_business_day(
        self,
    ):
        # type: (...) -> None

        total_blocking_days = self._get_total_blocking_days()
        days_open = []
        if self.pos_list is not None:
            for pos_id in sorted(self.pos_list):
                self._append_pos_if_must_block(
                    pos_id=str(pos_id),
                    days_open=days_open,
                    days_blocking=total_blocking_days,
                )

            if len(days_open) > 0:
                days_open = [item.get("pos") for item in sorted(days_open, key=lambda line: line['date'])]
                message = "$BLOCKING_DAYS_FOR_OPENING_BUSINESS_DAY|{0}|{1}".format(
                    total_blocking_days,
                    ", ".join(days_open),
                )

                show_messagebox(
                    posid=self.current_pos_id,
                    message=message,
                )

                raise StopAction()

    @staticmethod
    def _append_pos_if_must_block(
        pos_id,             # type: str
        days_open,          # type: List
        days_blocking,      # type: int
    ):
        # type: (...) -> None

        if pos_id == '0':
            return

        model = get_model(
            posid=pos_id,
        )
        if get_podtype(model) in ("TT", "DL"):
            return

        pos_is_open = model.find("PosState").get("state") not in ("CLOSED", "UNDEFINED")
        if not pos_is_open:
            return

        business_period = get_business_period(
            model=model,
        )
        business_period_date = datetime.strptime(business_period, '%Y%m%d')
        today_date = datetime.today()
        if (today_date-business_period_date).days > days_blocking:
            pos = translate_pos_label(model=model)
            days_open.append(
                {
                    "pos": "{}{}".format(pos, int(pos_id)),
                    "date": business_period,
                }
            )

    @staticmethod
    def _get_total_blocking_days(
    ):
        # type: (...) -> int

        global total_blocking_days_cache

        if total_blocking_days_cache is None:
            config = cfgtools.read(
                file_or_path=os.environ["LOADERCFG"],
            )

            total_blocking_days_cache = int(
                config.find_value(
                    path="Customizations.BlockingDaysForOpeningBusinessDay",
                ) or 3
            )

        return total_blocking_days_cache



import importlib
import logging
import os
from typing import Optional
from typing import (
    Optional,
    Dict,
)

# DO NOT REMOVE THIS - IT FIX A CYCLIC IMPORT
from application.domain import I18nImpl, I18n
from application.repository.impl import MwI18nRepository

from os.path import dirname, isfile, join
from msgbus import MBEasyContext

mb_context = None  # type: Optional[MBEasyContext]
logger = logging.getLogger("PosActions")

USECS_PER_SEC = 1000000
pos_config = None


class PosConfig(object):
    def __init__(
        self,
        max_transfer_value,                           # type: float
        skim_digit_limit,                             # type: Dict[str, int]
        sangria_levels,                               # type: str
        auto_generate_skim_number,                    # type: bool
        store_id,                                     # type: str
        min_value_to_ask_bordereau_justify_config,    # type: float
        ask_cash_out_justification,                   # type: bool
        keyboard_skim_numeric=True,                   # type: Optional[bool]
    ):

        self.max_transfer_value = max_transfer_value
        self.skim_digit_limit = skim_digit_limit
        self.sangria_levels = sangria_levels
        self.auto_generate_skim_number = auto_generate_skim_number
        self.store_id = store_id
        self.min_value_to_ask_bordereau_justify_config = min_value_to_ask_bordereau_justify_config
        self.keyboard_skim_numeric = keyboard_skim_numeric
        self.ask_cash_out_justification = ask_cash_out_justification


def get_modules(verify_file):
    mods = []
    for dir_path, dir_names, filenames in os.walk(dirname(__file__)):
        if dir_path == dirname(__file__):
            continue

        path_list = [join(dirname(__file__), "util"), join(dirname(__file__), "models")]
        if not verify_file(dir_path, path_list):
            continue

        for filename in filenames:
            if isfile(join(dir_path, filename)) and not filename.endswith("__init__.py") and filename.endswith(".py"):
                module_name = join(dir_path, filename)[len(dirname(__file__)) + 1:].replace(os.path.sep, '.')[:-3]
                mods.append("actions." + module_name)
    return mods

def verify_util_modules(path, path_list):
    return path in path_list

def verify_all_modules(path, path_list):
    return path not in path_list

def import_all_modules():
    for module in all_modules:
        importlib.import_module(module)

def import_util_modules():
    for module in util_modules:
        importlib.import_module(module)

def set_mb_context(mb_context_param):
    global mb_context
    mb_context = mb_context_param

def set_pos_config(pos_config_param):
    # type: (PosConfig) -> None

    global pos_config
    pos_config = pos_config_param

    for m in all_modules:
        reload(importlib.import_module(m))
    for m in util_modules:
        reload(importlib.import_module(m))


all_modules = get_modules(verify_all_modules)
all_modules = sorted(all_modules)

util_modules = get_modules(verify_util_modules)
util_modules = sorted(util_modules)

import_all_modules()
import_util_modules()



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
from io import BytesIO
import threading

import botocore.awsrequest
import botocore.session
from botocore import UNSIGNED
from botocore.config import Config
from botocore.compat import urlsplit
from botocore.exceptions import NoCredentialsError

import awscrt.http
from awscrt.s3 import S3Client, S3RequestType, S3RequestTlsMode
from awscrt.io import ClientBootstrap, DefaultHostResolver, EventLoopGroup
from awscrt.io import ClientTlsContext, TlsContextOptions
from awscrt.auth import AwsCredentialsProvider, AwsCredentials

from s3transfer.exceptions import TransferNotDoneError
from s3transfer.futures import BaseTransferFuture, BaseTransferMeta
from s3transfer.utils import CallArgs, OSUtils, get_callbacks
from s3transfer.constants import GB, MB

logger = logging.getLogger(__name__)


class CRTCredentialProviderAdapter:
    def __init__(self, botocore_credential_provider):
        self._botocore_credential_provider = botocore_credential_provider
        self._loaded_credentials = None
        self._lock = threading.Lock()

    def __call__(self):
        credentials = self._get_credentials().get_frozen_credentials()
        return AwsCredentials(credentials.access_key,
                              credentials.secret_key, credentials.token)

    def _get_credentials(self):
        with self._lock:
            if self._loaded_credentials is None:
                loaded_creds = self._botocore_credential_provider\
                    .load_credentials()
                if loaded_creds is None:
                    raise NoCredentialsError()
                self._loaded_credentials = loaded_creds
            return self._loaded_credentials


def create_s3_crt_client(region,
                         botocore_credential_provider=None,
                         num_threads=None,
                         target_throughput=5 * GB / 8,
                         part_size=8 * MB,
                         use_ssl=True,
                         verify=None):
    """
    :type region: str
    :param region: The region used for signing

    :type botocore_credential_provider:
        Optional[botocore.credentials.CredentialResolver]
    :param botocore_credential_provider: Provide credentials for CRT
        to sign the request if not set, the request will not be signed

    :type num_threads: Optional[int]
    :param num_threads: Number of worker threads generated. Default
        is the number of processors in the machine.

    :type target_throughput: Optional[int]
    :param target_throughput: Throughput target in Bytes.
        Default is 0.625 GB/s (which translates to 5 Gb/s).

    :type part_size: Optional[int]
    :param part_size: Size, in Bytes, of parts that files will be downloaded
        or uploaded in.

    :type use_ssl: boolean
    :param use_ssl: Whether or not to use SSL.  By default, SSL is used.
        Note that not all services support non-ssl connections.

    :type verify: Optional[boolean/string]
    :param verify: Whether or not to verify SSL certificates.
        By default SSL certificates are verified.  You can provide the
        following values:

        * False - do not validate SSL certificates.  SSL will still be
            used (unless use_ssl is False), but SSL certificates
            will not be verified.
        * path/to/cert/bundle.pem - A filename of the CA cert bundle to
            use. Specify this argument if you want to use a custom CA cert
            bundle instead of the default one on your system.
    """

    event_loop_group = EventLoopGroup(num_threads)
    host_resolver = DefaultHostResolver(event_loop_group)
    bootstrap = ClientBootstrap(event_loop_group, host_resolver)
    provider = None
    tls_connection_options = None

    tls_mode = S3RequestTlsMode.ENABLED if use_ssl \
        else S3RequestTlsMode.DISABLED
    if verify is not None:
        tls_ctx_options = TlsContextOptions()
        if verify:
            tls_ctx_options.override_default_trust_store_from_path(
                ca_filepath=verify)
        else:
            tls_ctx_options.verify_peer = False
        client_tls_option = ClientTlsContext(tls_ctx_options)
        tls_connection_options = client_tls_option.new_connection_options()
    if botocore_credential_provider:
        credentails_provider_adapter = CRTCredentialProviderAdapter(
            botocore_credential_provider)
        provider = AwsCredentialsProvider.new_delegate(
            credentails_provider_adapter)

    target_gbps = target_throughput * 8 / GB
    return S3Client(
        bootstrap=bootstrap,
        region=region,
        credential_provider=provider,
        part_size=part_size,
        tls_mode=tls_mode,
        tls_connection_options=tls_connection_options,
        throughput_target_gbps=target_gbps)


class CRTTransferManager(object):

    def __init__(self, crt_s3_client, crt_request_serializer, osutil=None):
        """A transfer manager interface for Amazon S3 on CRT s3 client.

        :type crt_s3_client: awscrt.s3.S3Client
        :param crt_s3_client: The CRT s3 client, handling all the
            HTTP requests and functions under then hood

        :type crt_request_serializer: s3transfer.crt.BaseCRTRequestSerializer
        :param crt_request_serializer: Serializer, generates unsigned crt HTTP
            request.

        :type osutil: s3transfer.utils.OSUtils
        :param osutil: OSUtils object to use for os-related behavior when
            using with transfer manager.
        """
        if osutil is None:
            self._osutil = OSUtils()
        self._crt_s3_client = crt_s3_client
        self._s3_args_creator = S3ClientArgsCreator(
            crt_request_serializer, self._osutil)
        self._future_coordinators = []
        self._semaphore = threading.Semaphore(128)  # not configurable
        # A counter to create unique id's for each transfer submitted.
        self._id_counter = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, *args):
        cancel = False
        if exc_type:
            cancel = True
        self._shutdown(cancel)

    def download(self, bucket, key, fileobj, extra_args=None,
                 subscribers=None):
        if extra_args is None:
            extra_args = {}
        if subscribers is None:
            subscribers = {}
        callargs = CallArgs(bucket=bucket, key=key, fileobj=fileobj,
                            extra_args=extra_args, subscribers=subscribers)
        return self._submit_transfer("get_object", callargs)

    def upload(self, fileobj, bucket, key, extra_args=None,
               subscribers=None):
        if extra_args is None:
            extra_args = {}
        if subscribers is None:
            subscribers = {}
        callargs = CallArgs(
            bucket=bucket, key=key, fileobj=fileobj,
            extra_args=extra_args, subscribers=subscribers)
        return self._submit_transfer("put_object", callargs)

    def delete(self, bucket, key, extra_args=None,
               subscribers=None):
        if extra_args is None:
            extra_args = {}
        if subscribers is None:
            subscribers = {}
        callargs = CallArgs(
            bucket=bucket, key=key, extra_args=extra_args,
            subscribers=subscribers)
        return self._submit_transfer("delete_object", callargs)

    def shutdown(self, cancel=False):
        self._shutdown(cancel)

    def _cancel_transfers(self):
        for coordinator in self._future_coordinators:
            if not coordinator.done():
                coordinator.cancel()

    def _finish_transfers(self):
        for coordinator in self._future_coordinators:
            coordinator.result()

    def _wait_transfers_done(self):
        for coordinator in self._future_coordinators:
            coordinator.wait_until_on_done_callbacks_complete()

    def _shutdown(self, cancel=False):
        if cancel:
            self._cancel_transfers()
        try:
            self._finish_transfers()

        except KeyboardInterrupt:
            self._cancel_transfers()
        except Exception:
            pass
        finally:
            self._wait_transfers_done()

    def _release_semaphore(self, **kwargs):
        self._semaphore.release()

    def _submit_transfer(self, request_type, call_args):
        on_done_after_calls = [self._release_semaphore]
        coordinator = CRTTransferCoordinator(transfer_id=self._id_counter)
        components = {
            'meta': CRTTransferMeta(self._id_counter, call_args),
            'coordinator': coordinator
        }
        future = CRTTransferFuture(**components)
        afterdone = AfterDoneHandler(coordinator)
        on_done_after_calls.append(afterdone)

        try:
            self._semaphore.acquire()
            on_queued = self._s3_args_creator.get_crt_callback(
                future, 'queued')
            on_queued()
            crt_callargs = self._s3_args_creator.get_make_request_args(
                request_type, call_args, coordinator,
                future, on_done_after_calls)
            crt_s3_request = self._crt_s3_client.make_request(**crt_callargs)
        except Exception as e:
            coordinator.set_exception(e, True)
            on_done = self._s3_args_creator.get_crt_callback(
                future, 'done', after_subscribers=on_done_after_calls)
            on_done(error=e)
        else:
            coordinator.set_s3_request(crt_s3_request)
        self._future_coordinators.append(coordinator)

        self._id_counter += 1
        return future


class CRTTransferMeta(BaseTransferMeta):
    """Holds metadata about the CRTTransferFuture"""

    def __init__(self, transfer_id=None, call_args=None):
        self._transfer_id = transfer_id
        self._call_args = call_args
        self._user_context = {}

    @property
    def call_args(self):
        return self._call_args

    @property
    def transfer_id(self):
        return self._transfer_id

    @property
    def user_context(self):
        return self._user_context


class CRTTransferFuture(BaseTransferFuture):
    def __init__(self, meta=None, coordinator=None):
        """The future associated to a submitted transfer request via CRT S3 client

        :type meta: s3transfer.crt.CRTTransferMeta
        :param meta: The metadata associated to the transfer future.

        :type coordinator: s3transfer.crt.CRTTransferCoordinator
        :param coordinator: The coordinator associated to the transfer future.
        """
        self._meta = meta
        if meta is None:
            self._meta = CRTTransferMeta()
        self._coordinator = coordinator

    @property
    def meta(self):
        return self._meta

    def done(self):
        return self._coordinator.done()

    def result(self, timeout=None):
        self._coordinator.result(timeout)

    def cancel(self):
        self._coordinator.cancel()

    def set_exception(self, exception):
        """Sets the exception on the future."""
        if not self.done():
            raise TransferNotDoneError(
                'set_exception can only be called once the transfer is '
                'complete.')
        self._coordinator.set_exception(exception, override=True)


class BaseCRTRequestSerializer:
    def serialize_http_request(self, transfer_type, future):
        """Serialize CRT HTTP requests.

        :type transfer_type: string
        :param transfer_type: the type of transfer made,
            e.g 'put_object', 'get_object', 'delete_object'

        :type future: s3transfer.crt.CRTTransferFuture

        :rtype: awscrt.http.HttpRequest
        :returns: An unsigned HTTP request to be used for the CRT S3 client
        """
        raise NotImplementedError('serialize_http_request()')


class BotocoreCRTRequestSerializer(BaseCRTRequestSerializer):
    def __init__(self, session, client_kwargs=None):
        """Serialize CRT HTTP request using botocore logic
        It also takes into account configuration from both the session
        and any keyword arguments that could be passed to
        `Session.create_client()` when serializing the request.

        :type session: botocore.session.Session

        :type client_kwargs: Optional[Dict[str, str]])
        :param client_kwargs: The kwargs for the botocore
            s3 client initialization.
        """
        self._session = session
        if client_kwargs is None:
            client_kwargs = {}
        self._resolve_client_config(session, client_kwargs)
        self._client = session.create_client(**client_kwargs)
        self._client.meta.events.register(
            'request-created.s3.*', self._capture_http_request)
        self._client.meta.events.register(
            'after-call.s3.*',
            self._change_response_to_serialized_http_request)
        self._client.meta.events.register(
            'before-send.s3.*', self._make_fake_http_response)

    def _resolve_client_config(self, session, client_kwargs):
        user_provided_config = None
        if session.get_default_client_config():
            user_provided_config = session.get_default_client_config()
        if 'config' in client_kwargs:
            user_provided_config = client_kwargs['config']

        client_config = Config(signature_version=UNSIGNED)
        if user_provided_config:
            client_config = user_provided_config.merge(client_config)
        client_kwargs['config'] = client_config
        client_kwargs["service_name"] = "s3"

    def _crt_request_from_aws_request(self, aws_request):
        url_parts = urlsplit(aws_request.url)
        crt_path = url_parts.path
        if url_parts.query:
            crt_path = '%s?%s' % (crt_path, url_parts.query)
        headers_list = []
        for name, value in aws_request.headers.items():
            if isinstance(value, str):
                headers_list.append((name, value))
            else:
                headers_list.append((name, str(value, 'utf-8')))

        crt_headers = awscrt.http.HttpHeaders(headers_list)
        # CRT requires body (if it exists) to be an I/O stream.
        crt_body_stream = None
        if aws_request.body:
            if hasattr(aws_request.body, 'seek'):
                crt_body_stream = aws_request.body
            else:
                crt_body_stream = BytesIO(aws_request.body)

        crt_request = awscrt.http.HttpRequest(
            method=aws_request.method,
            path=crt_path,
            headers=crt_headers,
            body_stream=crt_body_stream)
        return crt_request

    def _convert_to_crt_http_request(self, botocore_http_request):
        # Logic that does CRTUtils.crt_request_from_aws_request
        crt_request = self._crt_request_from_aws_request(
            botocore_http_request)
        if crt_request.headers.get("host") is None:
            # If host is not set, set it for the request before using CRT s3
            url_parts = urlsplit(botocore_http_request.url)
            crt_request.headers.set("host", url_parts.netloc)
        if crt_request.headers.get('Content-MD5') is not None:
            crt_request.headers.remove("Content-MD5")
        return crt_request

    def _capture_http_request(self, request, **kwargs):
        request.context['http_request'] = request

    def _change_response_to_serialized_http_request(
            self, context, parsed, **kwargs):
        request = context['http_request']
        parsed['HTTPRequest'] = request.prepare()

    def _make_fake_http_response(self, request, **kwargs):
        return botocore.awsrequest.AWSResponse(
            None,
            200,
            {},
            FakeRawResponse(b""),
        )

    def _get_botocore_http_request(self, client_method, call_args):
        return getattr(self._client, client_method)(
            Bucket=call_args.bucket, Key=call_args.key,
            **call_args.extra_args)['HTTPRequest']

    def serialize_http_request(self, transfer_type, future):
        botocore_http_request = self._get_botocore_http_request(
            transfer_type, future.meta.call_args)
        crt_request = self._convert_to_crt_http_request(botocore_http_request)
        return crt_request


class FakeRawResponse(BytesIO):
    def stream(self, amt=1024, decode_content=None):
        while True:
            chunk = self.read(amt)
            if not chunk:
                break
            yield chunk


class CRTTransferCoordinator:
    """A helper class for managing CRTTransferFuture"""

    def __init__(self, transfer_id=None, s3_request=None):
        self.transfer_id = transfer_id
        self._s3_request = s3_request
        self._lock = threading.Lock()
        self._exception = None
        self._crt_future = None
        self._done_event = threading.Event()

    @property
    def s3_request(self):
        return self._s3_request

    def set_done_callbacks_complete(self):
        self._done_event.set()

    def wait_until_on_done_callbacks_complete(self, timeout=None):
        self._done_event.wait(timeout)

    def set_exception(self, exception, override=False):
        with self._lock:
            if not self.done() or override:
                self._exception = exception

    def cancel(self):
        if self._s3_request:
            self._s3_request.cancel()

    def result(self, timeout=None):
        if self._exception:
            raise self._exception
        try:
            self._crt_future.result(timeout)
        except KeyboardInterrupt:
            self.cancel()
            raise
        finally:
            if self._s3_request:
                self._s3_request = None
            self._crt_future.result(timeout)

    def done(self):
        if self._crt_future is None:
            return False
        return self._crt_future.done()

    def set_s3_request(self, s3_request):
        self._s3_request = s3_request
        self._crt_future = self._s3_request.finished_future


class S3ClientArgsCreator:
    def __init__(self, crt_request_serializer, os_utils):
        self._request_serializer = crt_request_serializer
        self._os_utils = os_utils

    def get_make_request_args(
            self, request_type, call_args, coordinator, future,
            on_done_after_calls):
        recv_filepath = None
        send_filepath = None
        s3_meta_request_type = getattr(
            S3RequestType,
            request_type.upper(),
            S3RequestType.DEFAULT)
        on_done_before_calls = []
        if s3_meta_request_type == S3RequestType.GET_OBJECT:
            final_filepath = call_args.fileobj
            recv_filepath = self._os_utils.get_temp_filename(final_filepath)
            file_ondone_call = RenameTempFileHandler(
                coordinator, final_filepath, recv_filepath, self._os_utils)
            on_done_before_calls.append(file_ondone_call)
        elif s3_meta_request_type == S3RequestType.PUT_OBJECT:
            send_filepath = call_args.fileobj
            data_len = self._os_utils.get_file_size(send_filepath)
            call_args.extra_args["ContentLength"] = data_len

        crt_request = self._request_serializer.serialize_http_request(
            request_type, future)

        return {
            'request': crt_request,
            'type': s3_meta_request_type,
            'recv_filepath': recv_filepath,
            'send_filepath': send_filepath,
            'on_done': self.get_crt_callback(future, 'done',
                                             on_done_before_calls,
                                             on_done_after_calls),
            'on_progress': self.get_crt_callback(future, 'progress')
        }

    def get_crt_callback(self, future, callback_type,
                         before_subscribers=None, after_subscribers=None):

        def invoke_all_callbacks(*args, **kwargs):
            callbacks_list = []
            if before_subscribers is not None:
                callbacks_list += before_subscribers
            callbacks_list += get_callbacks(future, callback_type)
            if after_subscribers is not None:
                callbacks_list += after_subscribers
            for callback in callbacks_list:
                # The get_callbacks helper will set the first augment
                # by keyword, the other augments need to be set by keyword
                # as well
                if callback_type == "progress":
                    callback(bytes_transferred=args[0])
                else:
                    callback(*args, **kwargs)

        return invoke_all_callbacks


class RenameTempFileHandler:
    def __init__(self, coordinator, final_filename, temp_filename, osutil):
        self._coordinator = coordinator
        self._final_filename = final_filename
        self._temp_filename = temp_filename
        self._osutil = osutil

    def __call__(self, **kwargs):
        error = kwargs['error']
        if error:
            self._osutil.remove_file(self._temp_filename)
        else:
            try:
                self._osutil.rename_file(
                    self._temp_filename, self._final_filename)
            except Exception as e:
                self._osutil.remove_file(self._temp_filename)
                # the CRT future has done already at this point
                self._coordinator.set_exception(e)


class AfterDoneHandler:
    def __init__(self, coordinator):
        self._coordinator = coordinator

    def __call__(self, **kwargs):
        self._coordinator.set_done_callbacks_complete()



from abc import ABCMeta, abstractmethod

from production.model import ProductionOrder


class OrderCondition(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def evaluate(self, order):
        # type: (ProductionOrder) -> bool
        raise NotImplementedError()



from msgbus import MBEasyContext
from old_helper import round_half_away_from_zero
from pos_model import (
    Order,
    SaleItem,
)
from typing import (
    Dict,
    List,
)

from common import (
    CalculateTaxesItemCommon,
    FiscalParameterController,
    build_item_taxes,
)


class TipBuilderCommon(object):
    def __init__(self, tip_config, tip_config_range, mb_context, fiscal_parameter_controller, calculate_taxes_common):
        # type: (Dict, Dict, MBEasyContext, FiscalParameterController, CalculateTaxesItemCommon) -> None

        self.mb_context = mb_context
        self.tip_config = tip_config
        self.tip_config_range = tip_config_range
        self.fiscal_parameter_controller = fiscal_parameter_controller
        self.calculate_taxes_common = calculate_taxes_common

    def get_tip_items(self, order):
        # type: (Order) -> List[SaleItem]

        if order.tip:
            if len(self.tip_config_range) > 0:
                return self._build_tip_by_range(order)

            if len(self.tip_config) > 0 and self.tip_config['TipCode'] and self.tip_config['TipName']:
                price = order.tip
                return [self._get_sale_line_tip(self.tip_config['TipCode'], self.tip_config['TipName'].upper(), price)]

        return []

    def _build_tip_by_range(self, order):
        # type: (Order) -> List[SaleItem]

        items = []
        tip_ranges = self.get_percentage_by_tip(order.total_gross, order.tip)

        for key, value in tip_ranges.items():
            items.append(self._get_sale_line_tip(key, value.get("TipName"), value.get("TipValue")))

        return items

    def get_percentage_by_tip(self, total_gross, tip):
        # type: (float, float) -> Dict

        result = {}
        remaining_tip = tip
        for item in self.tip_config_range:
            code = item.get("TipCode")
            percentage = abs(item.get("PercentageUpper"))
            value = (percentage * total_gross) / 100.0
            current_tip = value if value <= remaining_tip else remaining_tip
            current_tip = round_half_away_from_zero(current_tip, 2)
            result[code] = {"TipName": item.get("TipName"), "TipValue": current_tip}
            remaining_tip -= current_tip

            if remaining_tip <= 0:
                break

        return self._check_and_adjust_tip_final_value(tip, result)

    def _check_and_adjust_tip_final_value(
            self,
            tip_value,      # type: float
            list_tip        # type: Dict
    ):
        # type:(...) -> Dict

        if len(list_tip) > 0:
            sum_added_tip = round_half_away_from_zero(sum([item.get("TipValue") for item in list_tip.values()]), 2)
            tip_difference = sum_added_tip != tip_value
            if abs(tip_difference) > 0.01:
                last_key_item = list_tip.keys()[-1]
                new_value = list_tip[last_key_item]['TipValue'] + (tip_value - sum_added_tip)
                list_tip[last_key_item]['TipValue'] = round_half_away_from_zero(new_value, 2)
                if list_tip[last_key_item]['TipValue'] <= 0:
                    list_tip.pop(last_key_item)

                list_tip = self._check_and_adjust_tip_final_value(tip_value, list_tip)

        return list_tip

    def _get_sale_line_tip(
        self,           # type: TipBuilderCommon
        part_code,      # type: str
        product_name,   # type: str
        price,          # type: float
    ):
        # type: (...) -> SaleItem

        sale_item = SaleItem()
        sale_item.part_code = str(part_code)
        sale_item.product_name = str(product_name.upper())
        sale_item.quantity = 1
        sale_item.unit_price = price
        sale_item.item_price = price
        sale_item.taxes = build_item_taxes(
            sale_item=sale_item,
            fiscal_parameter_controller=self.fiscal_parameter_controller,
            calculate_taxes_common=self.calculate_taxes_common,
        )
        return sale_item
